# pylint: disable=too-many-lines #noqa: PLR0915
import json
import re
import sys
import textwrap
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Mapping, Optional
from xml.dom.minidom import parseString

import pcs.lib.pacemaker.live as lib_pacemaker
import pcs.lib.resource_agent as lib_ra
from pcs import constraint, utils
from pcs.cli.cluster_property.output import PropertyConfigurationFacade
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    FUTURE_OPTION,
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    OUTPUT_FORMAT_VALUE_TEXT,
    Argv,
    InputModifiers,
    KeyValueParser,
    get_rule_str,
    group_by_keywords,
    wait_to_timeout,
)
from pcs.cli.common.tools import print_to_stderr, timeout_to_seconds_legacy
from pcs.cli.nvset import filter_out_expired_nvset, nvset_dto_list_to_lines
from pcs.cli.reports import process_library_reports
from pcs.cli.reports.output import error, warn
from pcs.cli.resource.common import check_is_not_stonith
from pcs.cli.resource.output import (
    operation_defaults_to_cmd,
    resource_agent_metadata_to_text,
    resource_defaults_to_cmd,
)
from pcs.cli.resource.parse_args import (
    parse_bundle_create_options,
    parse_bundle_reset_options,
    parse_bundle_update_options,
    parse_clone,
    parse_create_new,
    parse_create_old,
)
from pcs.cli.resource_agent import find_single_agent
from pcs.common import const, pacemaker, reports
from pcs.common.interface import dto
from pcs.common.pacemaker.defaults import CibDefaultsDto
from pcs.common.pacemaker.resource.operations import (
    OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME,
)
from pcs.common.str_tools import format_list_custom_last_separator
from pcs.lib.cib import const as cib_const
from pcs.lib.cib.resource import guest_node, primitive
from pcs.lib.cib.tools import get_resources
from pcs.lib.commands.resource import (
    _get_nodes_to_validate_against,
    _validate_guest_change,
)
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import is_true, validate_id
from pcs.settings import (
    pacemaker_wait_timeout_status as PACEMAKER_WAIT_TIMEOUT_STATUS,
)

if TYPE_CHECKING:
    from pcs.common.resource_agent.dto import ResourceAgentNameDto

RESOURCE_RELOCATE_CONSTRAINT_PREFIX = "pcs-relocate-"


def _detect_guest_change(
    meta_attributes: Mapping[str, str], allow_not_suitable_command: bool
) -> None:
    """
    Commandline options:
      * -f - CIB file
    """
    if not guest_node.is_node_name_in_options(meta_attributes):
        return

    env = utils.get_lib_env()
    cib = env.get_cib()
    (
        existing_nodes_names,
        existing_nodes_addrs,
        report_list,
    ) = _get_nodes_to_validate_against(env, cib)
    if env.report_processor.report_list(
        report_list
        + _validate_guest_change(
            cib,
            existing_nodes_names,
            existing_nodes_addrs,
            meta_attributes,
            allow_not_suitable_command,
            detect_remove=True,
        )
    ).has_errors:
        raise LibraryError()


def resource_utilization_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    resource_id = None
    if argv:
        resource_id = argv.pop(0)
        check_is_not_stonith(lib, [resource_id])
    utils.print_warning_if_utilization_attrs_has_no_effect(
        PropertyConfigurationFacade.from_properties_dtos(
            lib.cluster_property.get_properties(),
            lib.cluster_property.get_properties_metadata(),
        )
    )
    if not resource_id:
        print_resources_utilization()
        return
    if argv:
        set_resource_utilization(resource_id, argv)
    else:
        print_resource_utilization(resource_id)


def _defaults_set_create_cmd(
    lib_command: Callable[..., Any], argv: Argv, modifiers: InputModifiers
) -> None:
    modifiers.ensure_only_supported("-f", "--force")

    groups = group_by_keywords(
        argv, {"meta", "rule"}, implicit_first_keyword="options"
    )
    groups.ensure_unique_keywords()
    force_flags = set()
    if modifiers.get("--force"):
        force_flags.add(reports.codes.FORCE)

    lib_command(
        KeyValueParser(groups.get_args_flat("meta")).get_unique(),
        KeyValueParser(groups.get_args_flat("options")).get_unique(),
        nvset_rule=get_rule_str(groups.get_args_flat("rule")),
        force_flags=force_flags,
    )


def resource_defaults_set_create_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --force - allow unknown options
    """
    return _defaults_set_create_cmd(
        lib.cib_options.resource_defaults_create, argv, modifiers
    )


def resource_op_defaults_set_create_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --force - allow unknown options
    """
    return _defaults_set_create_cmd(
        lib.cib_options.operation_defaults_create, argv, modifiers
    )


def _filter_defaults(
    cib_defaults_dto: CibDefaultsDto, include_expired: bool
) -> CibDefaultsDto:
    return CibDefaultsDto(
        instance_attributes=(
            cib_defaults_dto.instance_attributes
            if include_expired
            else filter_out_expired_nvset(cib_defaults_dto.instance_attributes)
        ),
        meta_attributes=(
            cib_defaults_dto.meta_attributes
            if include_expired
            else filter_out_expired_nvset(cib_defaults_dto.meta_attributes)
        ),
    )


def _defaults_config_cmd(
    lib_command: Callable[[bool], CibDefaultsDto],
    defaults_to_cmd: Callable[[CibDefaultsDto], list[list[str]]],
    argv: Argv,
    modifiers: InputModifiers,
) -> None:
    """
    Options:
      * -f - CIB file
      * --all - display all nvsets including the ones with expired rules
      * --full - verbose output
      * --no-expire-check -- disable evaluating whether rules are expired
      * --output-format - supported formats: text, cmd, json
    """
    if argv:
        raise CmdLineInputError()
    modifiers.ensure_only_supported(
        "-f",
        "--all",
        "--full",
        "--no-expire-check",
        output_format_supported=True,
    )
    modifiers.ensure_not_mutually_exclusive("--all", "--no-expire-check")
    output_format = modifiers.get_output_format()
    if (
        modifiers.is_specified("--full")
        and output_format != OUTPUT_FORMAT_VALUE_TEXT
    ):
        raise CmdLineInputError(
            f"option '--full' is not compatible with '{output_format}' output format."
        )
    cib_defaults_dto = _filter_defaults(
        lib_command(not modifiers.get("--no-expire-check")),
        bool(modifiers.get("--all")),
    )
    if output_format == OUTPUT_FORMAT_VALUE_CMD:
        output = ";\n".join(
            " \\\n".join(cmd) for cmd in defaults_to_cmd(cib_defaults_dto)
        )
    elif output_format == OUTPUT_FORMAT_VALUE_JSON:
        output = json.dumps(dto.to_dict(cib_defaults_dto))
    else:
        output = "\n".join(
            nvset_dto_list_to_lines(
                cib_defaults_dto.meta_attributes,
                nvset_label="Meta Attrs",
                with_ids=bool(modifiers.get("--full")),
            )
        )
    if output:
        print(output)


def resource_defaults_config_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --full - verbose output
    """
    return _defaults_config_cmd(
        lib.cib_options.resource_defaults_config,
        resource_defaults_to_cmd,
        argv,
        modifiers,
    )


def resource_op_defaults_config_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --full - verbose output
    """
    return _defaults_config_cmd(
        lib.cib_options.operation_defaults_config,
        operation_defaults_to_cmd,
        argv,
        modifiers,
    )


def _defaults_set_remove_cmd(
    lib_command: Callable[..., Any], argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    lib_command(argv)


def resource_defaults_set_remove_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    return _defaults_set_remove_cmd(
        lib.cib_options.resource_defaults_remove, argv, modifiers
    )


def resource_op_defaults_set_remove_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    return _defaults_set_remove_cmd(
        lib.cib_options.operation_defaults_remove, argv, modifiers
    )


def _defaults_set_update_cmd(
    lib_command: Callable[..., Any], argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    set_id = argv[0]
    groups = group_by_keywords(argv[1:], {"meta"})
    groups.ensure_unique_keywords()
    lib_command(
        set_id, KeyValueParser(groups.get_args_flat("meta")).get_unique()
    )


def resource_defaults_set_update_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    return _defaults_set_update_cmd(
        lib.cib_options.resource_defaults_update, argv, modifiers
    )


def resource_op_defaults_set_update_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    return _defaults_set_update_cmd(
        lib.cib_options.operation_defaults_update, argv, modifiers
    )


def resource_defaults_update_cmd(
    lib: Any,
    argv: Argv,
    modifiers: InputModifiers,
) -> None:
    """
    Options:
      * -f - CIB file
    """
    del modifiers
    return lib.cib_options.resource_defaults_update(
        None, KeyValueParser(argv).get_unique()
    )


def resource_op_defaults_update_cmd(
    lib: Any,
    argv: Argv,
    modifiers: InputModifiers,
) -> None:
    """
    Options:
      * -f - CIB file
    """
    del modifiers
    return lib.cib_options.operation_defaults_update(
        None, KeyValueParser(argv).get_unique()
    )


def op_add_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --force - allow unknown options
    """
    if not argv:
        raise CmdLineInputError()
    check_is_not_stonith(lib, [argv[0]], "pcs stonith op add")
    resource_op_add(argv, modifiers)


def resource_op_add(argv: Argv, modifiers: InputModifiers) -> None:
    """
    Commandline options:
      * -f - CIB file
      * --force - allow unknown options
    """
    modifiers.ensure_only_supported("-f", "--force")
    if not argv:
        raise CmdLineInputError()
    res_id = argv.pop(0)

    # Check if we need to upgrade cib schema.
    # To do that, argv must be parsed, which is duplication of parsing in
    # resource_operation_add. But we need to upgrade the cib first before
    # calling that function. Hopefully, this will be fixed in the new pcs
    # architecture.

    # argv[0] is an operation name
    dom = None
    op_properties = utils.convert_args_to_tuples(argv[1:])
    for key, value in op_properties:
        if key == "on-fail" and value == "demote":
            dom = utils.cluster_upgrade_to_version(
                const.PCMK_ON_FAIL_DEMOTE_CIB_VERSION
            )
            break
    if dom is None:
        dom = utils.get_cib_dom()

    # add the requested operation
    utils.replace_cib_configuration(resource_operation_add(dom, res_id, argv))


def op_delete_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()
    resource_id = argv.pop(0)
    check_is_not_stonith(lib, [resource_id], "pcs stonith op delete")
    resource_operation_remove(resource_id, argv)


def parse_resource_options(
    argv: Argv,
) -> tuple[list[str], list[list[str]], list[str]]:
    """
    Commandline options: no options
    """
    ra_values = []
    op_values: list[list[str]] = []
    meta_values = []
    op_args = False
    meta_args = False
    for arg in argv:
        if arg == "op":
            op_args = True
            meta_args = False
            op_values.append([])
        elif arg == "meta":
            meta_args = True
            op_args = False
        elif op_args:
            if arg == "op":
                op_values.append([])
            elif "=" not in arg and op_values[-1]:
                op_values.append([])
                op_values[-1].append(arg)
            else:
                op_values[-1].append(arg)
        elif meta_args:
            if "=" in arg:
                meta_values.append(arg)
        else:
            ra_values.append(arg)
    return ra_values, op_values, meta_values


def resource_list_available(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --nodesc - don't display description
    """
    modifiers.ensure_only_supported("--nodesc")
    if len(argv) > 1:
        raise CmdLineInputError()

    search = argv[0] if argv else None
    agent_list = lib.resource_agent.list_agents(
        not modifiers.get("--nodesc"), search
    )

    if not agent_list:
        if search:
            utils.err("No resource agents matching the filter.")
        utils.err(
            "No resource agents available. Do you have resource agents installed?"
        )

    for agent_info in agent_list:
        name = agent_info["name"]
        shortdesc = agent_info["shortdesc"]
        if shortdesc:
            print(
                "{0} - {1}".format(
                    name,
                    _format_desc(
                        len(name + " - "), shortdesc.replace("\n", " ")
                    ),
                )
            )
        else:
            print(name)


def resource_list_options(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --full - show advanced
    """
    modifiers.ensure_only_supported("--full")
    if len(argv) != 1:
        raise CmdLineInputError()

    agent_name_str = argv[0]
    agent_name: ResourceAgentNameDto
    if ":" in agent_name_str:
        agent_name = lib.resource_agent.get_structured_agent_name(
            agent_name_str
        )
    else:
        agent_name = find_single_agent(
            lib.resource_agent.get_agents_list().names, agent_name_str
        )
    if agent_name.standard == "stonith":
        error(
            reports.messages.CommandArgumentTypeMismatch(
                "stonith / fence agents"
            ).message
            + " Please use 'pcs stonith describe' instead."
        )
    print(
        "\n".join(
            resource_agent_metadata_to_text(
                lib.resource_agent.get_agent_metadata(agent_name),
                lib.resource_agent.get_agent_default_operations(
                    agent_name
                ).operations,
                verbose=modifiers.is_specified("--full"),
            )
        )
    )


# Return the string formatted with a line length of terminal width  and indented
def _format_desc(indentation: int, desc: str) -> str:
    """
    Commandline options: no options
    """
    desc = " ".join(desc.split())
    dummy_rows, columns = utils.getTerminalSize()
    columns = max(int(columns), 40)
    afterindent = columns - indentation
    if afterindent < 1:
        afterindent = columns

    output = ""
    first = True
    for line in textwrap.wrap(desc, afterindent):
        if not first:
            output += " " * indentation
        output += line
        output += "\n"
        first = False

    return output.rstrip()


def resource_create(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:  # noqa: PLR0912
    """
    Options:
      * --agent-validation - use agent self validation of instance attributes
      * --before - specified resource inside a group before which new resource
        will be placed inside the group
      * --after - specified resource inside a group after which new resource
        will be placed inside the group
      * --group - specifies group in which resource will be created
      * --force - allow not existing agent, invalid operations or invalid
        instance attributes, allow not suitable command
      * --disabled - created resource will be disabled
      * --no-default-ops - do not add default operations
      * --wait
      * -f - CIB file
      * --future - enable future cli parser behavior
    """
    # pylint: disable=too-many-branches
    modifiers_deprecated = ["--before", "--after", "--group"]
    modifiers.ensure_only_supported(
        *(
            [
                "--agent-validation",
                "--force",
                "--disabled",
                "--no-default-ops",
                "--wait",
                "-f",
                FUTURE_OPTION,
            ]
            + ([] if modifiers.get(FUTURE_OPTION) else modifiers_deprecated)
        )
    )
    if len(argv) < 2:
        raise CmdLineInputError()

    ra_id = argv[0]
    ra_type = argv[1]

    if modifiers.get(FUTURE_OPTION):
        parts = parse_create_new(argv[2:])
    else:
        parts = parse_create_old(
            argv[2:], modifiers.get_subset(*modifiers_deprecated)
        )

    defined_options = set()
    if parts.bundle_id:
        defined_options.add("bundle")
    if parts.clone:
        defined_options.add("clone")
    if parts.promotable:
        defined_options.add("promotable")
    if parts.group:
        defined_options.add("group")
    if len(defined_options) > 1:
        raise CmdLineInputError(
            "you can specify only one of clone, promotable, bundle or {}group".format(
                "" if modifiers.get(FUTURE_OPTION) else "--"
            )
        )

    if parts.group:
        if parts.group.after_resource and parts.group.before_resource:
            raise CmdLineInputError(
                "you cannot specify both 'before' and 'after'"
                if modifiers.get(FUTURE_OPTION)
                else "you cannot specify both --before and --after"
            )

    if parts.promotable and "promotable" in parts.promotable.meta_attrs:
        raise CmdLineInputError(
            "you cannot specify both promotable option and promotable keyword"
        )

    settings = dict(
        allow_absent_agent=modifiers.get("--force"),
        allow_invalid_operation=modifiers.get("--force"),
        allow_invalid_instance_attributes=modifiers.get("--force"),
        ensure_disabled=modifiers.get("--disabled"),
        use_default_operations=not modifiers.get("--no-default-ops"),
        wait=modifiers.get("--wait"),
        allow_not_suitable_command=modifiers.get("--force"),
        enable_agent_self_validation=modifiers.get("--agent-validation"),
    )

    if parts.clone:
        lib.resource.create_as_clone(
            ra_id,
            ra_type,
            parts.primitive.operations,
            parts.primitive.meta_attrs,
            parts.primitive.instance_attrs,
            parts.clone.meta_attrs,
            clone_id=parts.clone.clone_id,
            allow_incompatible_clone_meta_attributes=modifiers.get("--force"),
            **settings,
        )
    elif parts.promotable:
        lib.resource.create_as_clone(
            ra_id,
            ra_type,
            parts.primitive.operations,
            parts.primitive.meta_attrs,
            parts.primitive.instance_attrs,
            dict(**parts.promotable.meta_attrs, promotable="true"),
            clone_id=parts.promotable.clone_id,
            allow_incompatible_clone_meta_attributes=modifiers.get("--force"),
            **settings,
        )
    elif parts.bundle_id:
        settings["allow_not_accessible_resource"] = modifiers.get("--force")
        lib.resource.create_into_bundle(
            ra_id,
            ra_type,
            parts.primitive.operations,
            parts.primitive.meta_attrs,
            parts.primitive.instance_attrs,
            parts.bundle_id,
            **settings,
        )
    elif parts.group:
        adjacent_resource_id = None
        put_after_adjacent = False
        if parts.group.after_resource:
            adjacent_resource_id = parts.group.after_resource
            put_after_adjacent = True
        if parts.group.before_resource:
            adjacent_resource_id = parts.group.before_resource
            put_after_adjacent = False

        lib.resource.create_in_group(
            ra_id,
            ra_type,
            parts.group.group_id,
            parts.primitive.operations,
            parts.primitive.meta_attrs,
            parts.primitive.instance_attrs,
            adjacent_resource_id=adjacent_resource_id,
            put_after_adjacent=put_after_adjacent,
            **settings,
        )
    else:
        lib.resource.create(
            ra_id,
            ra_type,
            parts.primitive.operations,
            parts.primitive.meta_attrs,
            parts.primitive.instance_attrs,
            **settings,
        )


def _parse_resource_move_ban(
    argv: Argv,
) -> tuple[str, Optional[str], Optional[str]]:
    resource_id = argv.pop(0)
    node = None
    lifetime = None
    while argv:
        arg = argv.pop(0)
        if arg.startswith("lifetime="):
            if lifetime:
                raise CmdLineInputError()
            lifetime = arg.split("=")[1]
            if lifetime and lifetime[0] in list("0123456789"):
                lifetime = "P" + lifetime
        elif not node:
            node = arg
        else:
            raise CmdLineInputError()
    return resource_id, node, lifetime


def resource_move_with_constraint(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --promoted
      * --wait
    """
    modifiers.ensure_only_supported("-f", "--promoted", "--wait")

    if not argv:
        raise CmdLineInputError("must specify a resource to move")
    if len(argv) > 3:
        raise CmdLineInputError()
    resource_id, node, lifetime = _parse_resource_move_ban(argv)

    lib.resource.move(
        resource_id,
        node=node,
        master=modifiers.is_specified("--promoted"),
        lifetime=lifetime,
        wait=modifiers.get("--wait"),
    )


def resource_move(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --promoted
      * --strict
      * --wait
    """
    modifiers.ensure_only_supported(
        "--promoted", "--strict", "--wait", hint_syntax_changed="0.12"
    )

    if not argv:
        raise CmdLineInputError("must specify a resource to move")
    resource_id = argv.pop(0)
    node = None
    if argv:
        node = argv.pop(0)
    if argv:
        raise CmdLineInputError()

    lib.resource.move_autoclean(
        resource_id,
        node=node,
        master=modifiers.is_specified("--promoted"),
        wait_timeout=wait_to_timeout(modifiers.get("--wait")),
        strict=modifiers.get("--strict"),
    )


def resource_ban(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --promoted
      * --wait
    """
    modifiers.ensure_only_supported("-f", "--promoted", "--wait")

    if not argv:
        raise CmdLineInputError("must specify a resource to ban")
    if len(argv) > 3:
        raise CmdLineInputError()
    resource_id, node, lifetime = _parse_resource_move_ban(argv)

    lib.resource.ban(
        resource_id,
        node=node,
        master=modifiers.is_specified("--promoted"),
        lifetime=lifetime,
        wait=modifiers.get("--wait"),
    )


def resource_unmove_unban(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --promoted
      * --wait
    """
    modifiers.ensure_only_supported("-f", "--expired", "--promoted", "--wait")

    if not argv:
        raise CmdLineInputError("must specify a resource to clear")
    if len(argv) > 2:
        raise CmdLineInputError()
    resource_id = argv.pop(0)
    node = argv.pop(0) if argv else None

    lib.resource.unmove_unban(
        resource_id,
        node=node,
        master=modifiers.is_specified("--promoted"),
        expired=modifiers.is_specified("--expired"),
        wait=modifiers.get("--wait"),
    )


def resource_standards(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()

    standards = lib.resource_agent.list_standards()

    if standards:
        print("\n".join(standards))
    else:
        utils.err("No standards found")


def resource_providers(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()

    providers = lib.resource_agent.list_ocf_providers()

    if providers:
        print("\n".join(providers))
    else:
        utils.err("No OCF providers found")


def resource_agents(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) > 1:
        raise CmdLineInputError()

    standard = argv[0] if argv else None

    agents = lib.resource_agent.list_agents_for_standard_and_provider(standard)

    if agents:
        print("\n".join(agents))
    else:
        utils.err(
            "No agents found{0}".format(
                " for {0}".format(argv[0]) if argv else ""
            )
        )


def update_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --agent-validation - use agent self validation of instance attributes
      * --wait
      * --force - allow invalid options, do not fail if not possible to get
        agent metadata, allow not suitable command
    """
    if not argv:
        raise CmdLineInputError()
    check_is_not_stonith(lib, [argv[0]], "pcs stonith update")
    resource_update(argv, modifiers)


# Update a resource, removing any args that are empty and adding/updating
# args that are not empty
def resource_update(args: Argv, modifiers: InputModifiers) -> None:  # noqa: PLR0912, PLR0915
    """
    Commandline options:
      * -f - CIB file
      * --agent-validation - use agent self validation of instance attributes
      * --wait
      * --force - allow invalid options, do not fail if not possible to get
        agent metadata, allow not suitable command
    """
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    modifiers.ensure_only_supported(
        "-f", "--wait", "--force", "--agent-validation"
    )
    if len(args) < 2:
        raise CmdLineInputError()
    res_id = args.pop(0)

    # Extract operation arguments
    ra_values, op_values, meta_values = parse_resource_options(args)

    wait = False
    wait_timeout = None
    if modifiers.is_specified("--wait"):
        # deprecated in the first version of 0.12
        process_library_reports(
            [
                reports.ReportItem.deprecation(
                    reports.messages.ResourceWaitDeprecated()
                )
            ]
        )

        wait_timeout = utils.validate_wait_get_timeout()
        wait = True

    # Check if we need to upgrade cib schema.
    # To do that, argv must be parsed, which is duplication of parsing below.
    # But we need to upgrade the cib first before calling that function.
    # Hopefully, this will be fixed in the new pcs architecture.

    cib_upgraded = False
    for op_argv in op_values:
        if cib_upgraded:
            break
        if len(op_argv) < 2:
            continue
        # argv[0] is an operation name
        op_vars = utils.convert_args_to_tuples(op_argv[1:])
        for key, value in op_vars:
            if key == "on-fail" and value == "demote":
                utils.cluster_upgrade_to_version(
                    const.PCMK_ON_FAIL_DEMOTE_CIB_VERSION
                )
                cib_upgraded = True
                break

    cib_xml = utils.get_cib()
    dom = utils.get_cib_dom(cib_xml=cib_xml)

    resource = utils.dom_get_resource(dom, res_id)
    if not resource:
        clone = utils.dom_get_clone(dom, res_id)
        master = utils.dom_get_master(dom, res_id)
        if clone or master:
            if master:
                clone = transform_master_to_clone(master)
            clone_child = utils.dom_elem_get_clone_ms_resource(clone)
            if clone_child:
                child_id = clone_child.getAttribute("id")
                new_args = ["meta"] + ra_values + meta_values
                for op_args in op_values:
                    if op_args:
                        new_args += ["op"] + op_args
                return resource_update_clone(
                    dom, clone, child_id, new_args, wait, wait_timeout
                )
        utils.err("Unable to find resource: %s" % res_id)

    params = utils.convert_args_to_tuples(ra_values)

    agent_name = _get_resource_agent_name_from_rsc_el(resource)
    try:
        agent_facade = _get_resource_agent_facade(agent_name)
        report_list = primitive.validate_resource_instance_attributes_update(
            utils.cmd_runner(),
            agent_facade,
            dict(params),
            res_id,
            get_resources(lib_pacemaker.get_cib(cib_xml)),
            force=bool(modifiers.get("--force")),
            enable_agent_self_validation=bool(
                modifiers.get("--agent-validation")
            ),
        )
        if report_list:
            process_library_reports(report_list)
    except lib_ra.ResourceAgentError as e:
        process_library_reports(
            [
                lib_ra.resource_agent_error_to_report_item(
                    e,
                    reports.get_severity(
                        reports.codes.FORCE, bool(modifiers.get("--force"))
                    ),
                )
            ]
        )

    utils.dom_update_instance_attr(resource, params)

    remote_node_name = utils.dom_get_resource_remote_node_name(resource)

    # The "remote-node" meta attribute makes sense (and causes creation of
    # inner pacemaker resource) only for primitive. The meta attribute
    # "remote-node" has no special meaning for clone/master. So there is no
    # need for checking this attribute in clone/master.
    #
    # It is ok to not to check it until this point in this function:
    # 1) Only master/clone element is updated if the parameter "res_id" is an id
    # of the clone/master element. In that case another function is called and
    # the code path does not reach this point.
    # 2) No persistent changes happened until this line if the parameter
    # "res_id" is an id of the primitive.
    meta_options = KeyValueParser(meta_values).get_unique()
    if remote_node_name != guest_node.get_guest_option_value(meta_options):
        _detect_guest_change(
            meta_options,
            bool(modifiers.get("--force")),
        )

    # TODO: validation should be added after migrating the command to the new
    # architecture
    if any(meta_options.values()):
        command = "stonith" if agent_name.is_stonith else "resource"
        warn(
            "Meta attributes are not validated by this command. For "
            f"validation, please use 'pcs {command} meta' instead."
        )

    utils.dom_update_meta_attr(
        resource, utils.convert_args_to_tuples(meta_values)
    )

    operations = resource.getElementsByTagName("operations")
    if not operations:
        operations = dom.createElement("operations")
        resource.appendChild(operations)
    else:
        operations = operations[0]

    get_role = partial(
        pacemaker.role.get_value_for_cib,
        is_latest_supported=utils.isCibVersionSatisfied(
            dom, const.PCMK_NEW_ROLES_CIB_VERSION
        ),
    )
    for op_argv in op_values:
        if not op_argv:
            continue

        op_name = op_argv[0]
        if op_name.find("=") != -1:
            utils.err(
                "%s does not appear to be a valid operation action" % op_name
            )

        if len(op_argv) < 2:
            continue

        op_role = ""
        op_vars = utils.convert_args_to_tuples(op_argv[1:])

        for key, value in op_vars:
            if key == "role":
                op_role = get_role(value)
                break

        updating_op = None
        updating_op_before = None
        for existing_op in operations.getElementsByTagName("op"):
            if updating_op:
                updating_op_before = existing_op
                break
            existing_op_name = existing_op.getAttribute("name")
            existing_op_role = get_role(existing_op.getAttribute("role"))
            if existing_op_role == op_role and existing_op_name == op_name:
                updating_op = existing_op
                continue

        if updating_op:
            updating_op.parentNode.removeChild(updating_op)
        dom = resource_operation_add(
            dom,
            res_id,
            op_argv,
            validate_strict=False,
            before_op=updating_op_before,
        )

    utils.replace_cib_configuration(dom)

    if (
        remote_node_name
        and remote_node_name
        != utils.dom_get_resource_remote_node_name(resource)
    ):
        # if the resource was a remote node and it is not anymore, (or its name
        # changed) we need to tell pacemaker about it
        output, retval = utils.run(
            ["crm_node", "--force", "--remove", remote_node_name]
        )

    if modifiers.is_specified("--wait"):
        args = ["crm_resource", "--wait"]
        if wait_timeout:
            args.extend(["--timeout=%s" % wait_timeout])
        output, retval = utils.run(args)
        running_on = utils.resource_running_on(res_id)
        if retval == 0:
            print_to_stderr(running_on["message"])
        else:
            msg = []
            if retval == PACEMAKER_WAIT_TIMEOUT_STATUS:
                msg.append("waiting timeout")
            msg.append(running_on["message"])
            if retval != 0 and output:
                msg.append("\n" + output)
            utils.err("\n".join(msg).strip())
    return None


def resource_update_clone(dom, clone, res_id, args, wait, wait_timeout):
    """
    Commandline options:
      * -f - CIB file
    """
    dom, dummy_clone_id = resource_clone_create(
        dom, [res_id] + args, update_existing=True
    )

    utils.replace_cib_configuration(dom)

    if wait:
        args = ["crm_resource", "--wait"]
        if wait_timeout:
            args.extend(["--timeout=%s" % wait_timeout])
        output, retval = utils.run(args)
        running_on = utils.resource_running_on(clone.getAttribute("id"))
        if retval == 0:
            print_to_stderr(running_on["message"])
        else:
            msg = []
            if retval == PACEMAKER_WAIT_TIMEOUT_STATUS:
                msg.append("waiting timeout")
            msg.append(running_on["message"])
            if retval != 0 and output:
                msg.append("\n" + output)
            utils.err("\n".join(msg).strip())

    return dom


def transform_master_to_clone(master_element):
    # create a new clone element with the same id
    dom = master_element.ownerDocument
    clone_element = dom.createElement("clone")
    clone_element.setAttribute("id", master_element.getAttribute("id"))
    # place it next to the master element
    master_element.parentNode.insertBefore(clone_element, master_element)
    # move all master's children to the clone
    while master_element.firstChild:
        clone_element.appendChild(master_element.firstChild)
    # remove the master
    master_element.parentNode.removeChild(master_element)
    # set meta to make the clone promotable
    utils.dom_update_meta_attr(clone_element, [("promotable", "true")])
    return clone_element


def resource_operation_add(  # noqa: PLR0912, PLR0915
    dom, res_id, argv, validate_strict=True, before_op=None
):
    """
    Commandline options:
      * --force
    """
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    if not argv:
        raise CmdLineInputError()

    res_el = utils.dom_get_resource(dom, res_id)
    if not res_el:
        utils.err("Unable to find resource: %s" % res_id)

    op_name = argv.pop(0)
    op_properties = utils.convert_args_to_tuples(argv)

    if "=" in op_name:
        utils.err("%s does not appear to be a valid operation action" % op_name)
    if "--force" not in utils.pcs_options:
        valid_attrs = [
            "id",
            "name",
            "interval",
            "description",
            "start-delay",
            "interval-origin",
            "timeout",
            "enabled",
            "record-pending",
            "role",
            "on-fail",
            OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME,
        ]
        for key, value in op_properties:
            if key not in valid_attrs:
                utils.err(
                    "%s is not a valid op option (use --force to override)"
                    % key
                )
            if key == "role" and value not in const.PCMK_ROLES:
                utils.err(
                    "role must be: {} (use --force to override)".format(
                        format_list_custom_last_separator(
                            const.PCMK_ROLES, " or "
                        )
                    )
                )

    interval = None
    for key, val in op_properties:
        if key == "interval":
            interval = val
            break
    if not interval:
        interval = "60s" if op_name == "monitor" else "0s"
        op_properties.append(("interval", interval))

    op_properties.sort(key=lambda a: a[0])
    op_properties.insert(0, ("name", op_name))

    generate_id = True
    for name, value in op_properties:
        if name == "id":
            op_id = value
            generate_id = False
            id_valid, id_error = utils.validate_xml_id(value, "operation id")
            if not id_valid:
                utils.err(id_error)
            if utils.does_id_exist(dom, value):
                utils.err(
                    "id '%s' is already in use, please specify another one"
                    % value
                )
    if generate_id:
        op_id = "%s-%s-interval-%s" % (res_id, op_name, interval)
        op_id = utils.find_unique_id(dom, op_id)

    op_el = dom.createElement("op")
    op_el.setAttribute("id", op_id)
    for key, val in op_properties:
        if key == OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME:
            attrib_el = dom.createElement("instance_attributes")
            attrib_el.setAttribute(
                "id", utils.find_unique_id(dom, "params-" + op_id)
            )
            op_el.appendChild(attrib_el)
            nvpair_el = dom.createElement("nvpair")
            nvpair_el.setAttribute("name", key)
            nvpair_el.setAttribute("value", val)
            nvpair_el.setAttribute(
                "id", utils.find_unique_id(dom, "-".join((op_id, key, val)))
            )
            attrib_el.appendChild(nvpair_el)
        elif key == "role" and "--force" not in utils.pcs_options:
            op_el.setAttribute(
                key,
                pacemaker.role.get_value_for_cib(
                    val,
                    utils.isCibVersionSatisfied(
                        dom, const.PCMK_NEW_ROLES_CIB_VERSION
                    ),
                ),
            )
        else:
            op_el.setAttribute(key, val)

    operations = res_el.getElementsByTagName("operations")
    if not operations:
        operations = dom.createElement("operations")
        res_el.appendChild(operations)
    else:
        operations = operations[0]
        duplicate_op_list = utils.operation_exists(operations, op_el)
        if duplicate_op_list:
            utils.err(
                "operation %s with interval %ss already specified for %s:\n%s"
                % (
                    op_el.getAttribute("name"),
                    timeout_to_seconds_legacy(op_el.getAttribute("interval")),
                    res_id,
                    "\n".join(
                        [operation_to_string(op) for op in duplicate_op_list]
                    ),
                )
            )
        if validate_strict and "--force" not in utils.pcs_options:
            duplicate_op_list = utils.operation_exists_by_name(
                operations, op_el
            )
            if duplicate_op_list:
                msg = (
                    "operation {action} already specified for {res}"
                    + ", use --force to override:\n{op}"
                )
                utils.err(
                    msg.format(
                        action=op_el.getAttribute("name"),
                        res=res_id,
                        op="\n".join(
                            [
                                operation_to_string(op)
                                for op in duplicate_op_list
                            ]
                        ),
                    )
                )

    operations.insertBefore(op_el, before_op)
    return dom


def resource_operation_remove(res_id: str, argv: Argv) -> None:  # noqa: PLR0912
    """
    Commandline options:
      * -f - CIB file
    """
    # pylint: disable=too-many-branches
    # if no args, then we're removing an operation id

    # Do not ever remove an operations element, even if it is empty. There may
    # be ACLs set in pacemaker which allow "write" for op elements (adding,
    # changing and removing) but not operations elements. In such a case,
    # removing an operations element would cause the whole change to be
    # rejected by pacemaker with a "permission denied" message.
    # https://bugzilla.redhat.com/show_bug.cgi?id=1642514

    dom = utils.get_cib_dom()
    if not argv:
        for operation in dom.getElementsByTagName("op"):
            if operation.getAttribute("id") == res_id:
                parent = operation.parentNode
                parent.removeChild(operation)
                utils.replace_cib_configuration(dom)
                return
        utils.err("unable to find operation id: %s" % res_id)

    original_argv = " ".join(argv)

    op_name = argv.pop(0)
    resource_el = None

    for resource in dom.getElementsByTagName("primitive"):
        if resource.getAttribute("id") == res_id:
            resource_el = resource
            break

    if not resource_el:
        utils.err("Unable to find resource: %s" % res_id)
        # return to let mypy know that resource_el is not None anymore
        return

    remove_all = False
    if not argv:
        remove_all = True

    op_properties = utils.convert_args_to_tuples(argv)
    op_properties.append(("name", op_name))
    found_match = False
    for op in resource_el.getElementsByTagName("op"):
        temp_properties = []
        for attr_name in op.attributes.keys():  # noqa: SIM118, attributes is not a dict
            if attr_name == "id":
                continue
            temp_properties.append(
                (attr_name, op.attributes.get(attr_name).nodeValue)
            )

        if remove_all and op.attributes["name"].value == op_name:
            found_match = True
            parent = op.parentNode
            parent.removeChild(op)
        elif not set(op_properties) ^ set(temp_properties):
            found_match = True
            parent = op.parentNode
            parent.removeChild(op)
            break

    if not found_match:
        utils.err("Unable to find operation matching: %s" % original_argv)

    utils.replace_cib_configuration(dom)


def resource_group_rm_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --wait
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("--wait", "-f")
    if not argv:
        raise CmdLineInputError()
    group_name = argv.pop(0)
    resource_ids = argv

    cib_dom = resource_group_rm(utils.get_cib_dom(), group_name, resource_ids)

    if modifiers.is_specified("--wait"):
        # deprecated in the first version of 0.12
        process_library_reports(
            [
                reports.ReportItem.deprecation(
                    reports.messages.ResourceWaitDeprecated()
                )
            ]
        )

        wait_timeout = utils.validate_wait_get_timeout()

    utils.replace_cib_configuration(cib_dom)

    if modifiers.is_specified("--wait"):
        args = ["crm_resource", "--wait"]
        if wait_timeout:
            args.extend(["--timeout=%s" % wait_timeout])
        output, retval = utils.run(args)
        if retval != 0:
            msg = []
            if retval == PACEMAKER_WAIT_TIMEOUT_STATUS:
                msg.append("waiting timeout")
            if output:
                msg.append("\n" + output)
            utils.err("\n".join(msg).strip())


def resource_group_add_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --wait
      * -f - CIB file
      * --after - place a resource in a group after the specified resource in
        the group
      * --before - place a resource in a group before the specified resource in
        the group
    """
    modifiers.ensure_only_supported("--wait", "-f", "--after", "--before")
    if len(argv) < 2:
        raise CmdLineInputError()

    group_name = argv.pop(0)
    resource_names = argv
    adjacent_name = None
    after_adjacent = True
    if modifiers.is_specified("--after") and modifiers.is_specified("--before"):
        raise CmdLineInputError("you cannot specify both --before and --after")
    if modifiers.is_specified("--after"):
        adjacent_name = modifiers.get("--after")
        after_adjacent = True
    elif modifiers.is_specified("--before"):
        adjacent_name = modifiers.get("--before")
        after_adjacent = False

    lib.resource.group_add(
        group_name,
        resource_names,
        adjacent_resource_id=adjacent_name,
        put_after_adjacent=after_adjacent,
        wait=modifiers.get("--wait"),
    )


def resource_clone(
    lib: Any, argv: Argv, modifiers: InputModifiers, promotable: bool = False
) -> None:
    """
    Options:
      * --wait
      * -f - CIB file
      * --force - allow to clone stonith resource
    """
    modifiers.ensure_only_supported("-f", "--force", "--wait")
    if not argv:
        raise CmdLineInputError()

    res = argv[0]
    check_is_not_stonith(lib, [res])
    cib_dom = utils.get_cib_dom()

    if modifiers.is_specified("--wait"):
        # deprecated in the first version of 0.12
        process_library_reports(
            [
                reports.ReportItem.deprecation(
                    reports.messages.ResourceWaitDeprecated()
                )
            ]
        )

        wait_timeout = utils.validate_wait_get_timeout()

    force_flags = set()
    if modifiers.get("--force"):
        force_flags.add(reports.codes.FORCE)

    cib_dom, clone_id = resource_clone_create(
        cib_dom, argv, promotable=promotable, force_flags=force_flags
    )
    cib_dom = constraint.constraint_resource_update(res, cib_dom)
    utils.replace_cib_configuration(cib_dom)

    if modifiers.is_specified("--wait"):
        args = ["crm_resource", "--wait"]
        if wait_timeout:
            args.extend(["--timeout=%s" % wait_timeout])
        output, retval = utils.run(args)
        running_on = utils.resource_running_on(clone_id)
        if retval == 0:
            print_to_stderr(running_on["message"])
        else:
            msg = []
            if retval == PACEMAKER_WAIT_TIMEOUT_STATUS:
                msg.append("waiting timeout")
            msg.append(running_on["message"])
            if output:
                msg.append("\n" + output)
            utils.err("\n".join(msg).strip())


def _resource_is_ocf(resource_el) -> bool:
    return resource_el.getAttribute("class") == "ocf"


def _get_resource_agent_name_from_rsc_el(
    resource_el,
) -> lib_ra.ResourceAgentName:
    return lib_ra.ResourceAgentName(
        resource_el.getAttribute("class"),
        resource_el.getAttribute("provider"),
        resource_el.getAttribute("type"),
    )


def _get_resource_agent_facade(
    resource_agent: lib_ra.ResourceAgentName,
) -> lib_ra.ResourceAgentFacade:
    return lib_ra.ResourceAgentFacadeFactory(
        utils.cmd_runner(), utils.get_report_processor()
    ).facade_from_parsed_name(resource_agent)


def resource_clone_create(  # noqa: PLR0912
    cib_dom, argv, update_existing=False, promotable=False, force_flags=()
):
    """
    Commandline options:
      * --force - allow to clone stonith resource
    """
    # pylint: disable=too-many-branches
    name = argv.pop(0)

    resources_el = cib_dom.getElementsByTagName("resources")[0]
    element = utils.dom_get_resource(resources_el, name) or utils.dom_get_group(
        resources_el, name
    )
    if not element:
        utils.err("unable to find group or resource: %s" % name)

    if element.parentNode.tagName == "bundle":
        utils.err("cannot clone bundle resource")

    if not update_existing:
        if utils.dom_get_resource_clone(
            cib_dom, name
        ) or utils.dom_get_resource_masterslave(cib_dom, name):
            utils.err("%s is already a clone resource" % name)

        if utils.dom_get_group_clone(
            cib_dom, name
        ) or utils.dom_get_group_masterslave(cib_dom, name):
            utils.err("cannot clone a group that has already been cloned")
    else:
        if element.parentNode.tagName != "clone":
            utils.err("%s is not currently a clone" % name)
        clone = element.parentNode

    # If element is currently in a group and it's the last member, we get rid
    # of the group
    if (
        element.parentNode.tagName == "group"
        and element.parentNode.getElementsByTagName("primitive").length <= 1
    ):
        element.parentNode.parentNode.removeChild(element.parentNode)

    if element.getAttribute("class") == "stonith":
        process_library_reports(
            [
                reports.ReportItem(
                    severity=reports.item.get_severity(
                        reports.codes.FORCE,
                        is_forced=reports.codes.FORCE in force_flags,
                    ),
                    message=reports.messages.CloningStonithResourcesHasNoEffect(
                        [name]
                    ),
                )
            ]
        )

    parts = parse_clone(argv, promotable=promotable)
    _check_clone_incompatible_options_child(
        element, parts.meta_attrs, force=reports.codes.FORCE in force_flags
    )

    if not update_existing:
        clone_id = parts.clone_id
        if clone_id is not None:
            report_list = []
            validate_id(clone_id, reporter=report_list)
            if report_list:
                raise CmdLineInputError("invalid id '{}'".format(clone_id))
            if utils.does_id_exist(cib_dom, clone_id):
                raise CmdLineInputError(
                    "id '{}' already exists".format(clone_id),
                )
        else:
            clone_id = utils.find_unique_id(cib_dom, name + "-clone")
        clone = cib_dom.createElement("clone")
        clone.setAttribute("id", clone_id)
        clone.appendChild(element)
        resources_el.appendChild(clone)

    # TODO: validation should be added after migrating the command to the new
    # architecture
    if any(
        value
        for name, value in parts.meta_attrs.items()
        if name != "promotable" or not promotable
    ):
        warn(
            reports.messages.MetaAttrsNotValidatedUnsupportedType(
                [cib_const.TAG_RESOURCE_CLONE]
            ).message
        )
    utils.dom_update_meta_attr(clone, sorted(parts.meta_attrs.items()))

    return cib_dom, clone.getAttribute("id")


def _check_clone_incompatible_options_child(
    child_el,
    clone_meta_attrs: Mapping[str, str],
    force: bool = False,
):
    report_list = []
    if child_el.tagName == "primitive":
        report_list = _check_clone_incompatible_options_primitive(
            child_el, clone_meta_attrs, force=force
        )
    elif child_el.tagName == "group":
        group_id = child_el.getAttribute("id")
        for primitive_el in utils.get_group_children_el_from_el(child_el):
            report_list.extend(
                _check_clone_incompatible_options_primitive(
                    primitive_el,
                    clone_meta_attrs,
                    group_id=group_id,
                    force=force,
                )
            )
    if report_list:
        process_library_reports(report_list)


def _check_clone_incompatible_options_primitive(
    primitive_el,
    clone_meta_attrs: Mapping[str, str],
    group_id: Optional[str] = None,
    force: bool = False,
) -> reports.ReportItemList:
    resource_agent_name = _get_resource_agent_name_from_rsc_el(primitive_el)
    primitive_id = primitive_el.getAttribute("id")
    if not _resource_is_ocf(primitive_el):
        for incompatible_attribute in ("globally-unique", "promotable"):
            if is_true(clone_meta_attrs.get(incompatible_attribute, "0")):
                return [
                    reports.ReportItem.error(
                        reports.messages.ResourceCloneIncompatibleMetaAttributes(
                            incompatible_attribute,
                            resource_agent_name.to_dto(),
                            resource_id=primitive_id,
                            group_id=group_id,
                        )
                    )
                ]
    else:
        try:
            resource_agent_facade = _get_resource_agent_facade(
                resource_agent_name
            )
        except lib_ra.ResourceAgentError as e:
            return [
                lib_ra.resource_agent_error_to_report_item(
                    e, reports.get_severity(reports.codes.FORCE, force)
                )
            ]
        if resource_agent_facade.metadata.ocf_version == "1.1" and (
            is_true(clone_meta_attrs.get("promotable", "0"))
            and not resource_agent_facade.metadata.provides_promotability
        ):
            return [
                reports.ReportItem(
                    reports.get_severity(reports.codes.FORCE, force),
                    reports.messages.ResourceCloneIncompatibleMetaAttributes(
                        "promotable",
                        resource_agent_name.to_dto(),
                        resource_id=primitive_id,
                        group_id=group_id,
                    ),
                )
            ]
    return []


def resource_clone_master_remove(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --wait
    """
    # pylint: disable=too-many-locals
    del lib
    modifiers.ensure_only_supported("-f", "--wait")
    if len(argv) != 1:
        raise CmdLineInputError()

    name = argv.pop()
    dom = utils.get_cib_dom()
    resources_el = dom.documentElement.getElementsByTagName("resources")[0]

    # get the resource no matter if user entered a clone or a cloned resource
    resource = (
        utils.dom_get_resource(resources_el, name)
        or utils.dom_get_group(resources_el, name)
        or utils.dom_get_clone_ms_resource(resources_el, name)
    )
    if not resource:
        utils.err("could not find resource: %s" % name)
    resource_id = resource.getAttribute("id")
    clone = utils.dom_get_resource_clone_ms_parent(resources_el, resource_id)
    if not clone:
        utils.err("'%s' is not a clone resource" % name)

    if modifiers.is_specified("--wait"):
        # deprecated in the first version of 0.12
        process_library_reports(
            [
                reports.ReportItem.deprecation(
                    reports.messages.ResourceWaitDeprecated()
                )
            ]
        )

        wait_timeout = utils.validate_wait_get_timeout()

    # if user requested uncloning a resource contained in a cloned group
    # remove the resource from the group and leave the clone itself alone
    # unless the resource is the last one in the group
    clone_child = utils.dom_get_clone_ms_resource(
        resources_el, clone.getAttribute("id")
    )
    if (
        clone_child.tagName == "group"
        and resource.tagName != "group"
        and len(clone_child.getElementsByTagName("primitive")) > 1
    ):
        resource_group_rm(dom, clone_child.getAttribute("id"), [resource_id])
    else:
        remove_resource_references(dom, clone.getAttribute("id"))
        clone.parentNode.appendChild(resource)
        clone.parentNode.removeChild(clone)
    utils.replace_cib_configuration(dom)

    if modifiers.is_specified("--wait"):
        args = ["crm_resource", "--wait"]
        if wait_timeout:
            args.extend(["--timeout=%s" % wait_timeout])
        output, retval = utils.run(args)
        running_on = utils.resource_running_on(resource_id)
        if retval == 0:
            print_to_stderr(running_on["message"])
        else:
            msg = []
            if retval == PACEMAKER_WAIT_TIMEOUT_STATUS:
                msg.append("waiting timeout")
            msg.append(running_on["message"])
            if output:
                msg.append("\n" + output)
            utils.err("\n".join(msg).strip())


def stonith_level_rm_device(cib_dom, stn_id):
    """
    Commandline options: no options
    """
    topology_el_list = cib_dom.getElementsByTagName("fencing-topology")
    if not topology_el_list:
        return cib_dom
    topology_el = topology_el_list[0]
    for level_el in topology_el.getElementsByTagName("fencing-level"):
        device_list = level_el.getAttribute("devices").split(",")
        if stn_id in device_list:
            new_device_list = [dev for dev in device_list if dev != stn_id]
            if new_device_list:
                level_el.setAttribute("devices", ",".join(new_device_list))
            else:
                level_el.parentNode.removeChild(level_el)
    if not topology_el.getElementsByTagName("fencing-level"):
        topology_el.parentNode.removeChild(topology_el)
    return cib_dom


def remove_resource_references(
    dom, resource_id, output=False, constraints_element=None
):
    """
    Commandline options: no options
    NOTE: -f - will be used only if dom will be None
    """
    for obj_ref in dom.getElementsByTagName("obj_ref"):
        if obj_ref.getAttribute("id") == resource_id:
            tag = obj_ref.parentNode
            tag.removeChild(obj_ref)
            if tag.getElementsByTagName("obj_ref").length == 0:
                remove_resource_references(
                    dom,
                    tag.getAttribute("id"),
                    output=output,
                )
                tag.parentNode.removeChild(tag)
    constraint.remove_constraints_containing(
        resource_id, output, constraints_element, dom
    )
    stonith_level_rm_device(dom, resource_id)

    for permission in dom.getElementsByTagName("acl_permission"):
        if permission.getAttribute("reference") == resource_id:
            permission.parentNode.removeChild(permission)

    return dom


# This removes a resource from a group, but keeps it in the config
def resource_group_rm(cib_dom, group_name, resource_ids):
    """
    Commandline options: no options
    """
    dom = cib_dom.getElementsByTagName("configuration")[0]

    all_resources = len(resource_ids) == 0

    group_match = utils.dom_get_group(dom, group_name)
    if not group_match:
        utils.err("Group '%s' does not exist" % group_name)

    resources_to_move = []
    if all_resources:
        resources_to_move.extend(
            list(group_match.getElementsByTagName("primitive"))
        )
    else:
        for resource_id in resource_ids:
            resource = utils.dom_get_resource(group_match, resource_id)
            if resource:
                resources_to_move.append(resource)
            else:
                utils.err(
                    "Resource '%s' does not exist in group '%s'"
                    % (resource_id, group_name)
                )

    # If the group is in a clone, we don't delete the clone as there may be
    # constraints associated with it which the user may want to keep. However,
    # there may be several resources in the group. In that case there is no way
    # to figure out which one of them should stay in the clone. So we forbid
    # removing all resources from a cloned group unless there is just one
    # resource.
    # This creates an inconsistency:
    # - consider a cloned group with two resources
    # - move one resource from the group - it becomes a primitive
    # - move the last resource from the group - it stays in the clone
    # So far there has been no request to change this behavior. Unless there is
    # a request / reason to change it, we'll keep it that way.
    is_cloned_group = group_match.parentNode.tagName in ["clone", "master"]
    res_in_group = len(group_match.getElementsByTagName("primitive"))
    if (
        is_cloned_group
        and res_in_group > 1
        and len(resources_to_move) == res_in_group
    ):
        utils.err("Cannot remove all resources from a cloned group")
    target_node = group_match.parentNode
    if is_cloned_group and res_in_group > 1:
        target_node = dom.getElementsByTagName("resources")[0]
    for resource in resources_to_move:
        resource.parentNode.removeChild(resource)
        target_node.appendChild(resource)

    if not group_match.getElementsByTagName("primitive"):
        group_match.parentNode.removeChild(group_match)
        remove_resource_references(dom, group_name, output=True)

    return cib_dom


def resource_group_list(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("-f")
    if argv:
        raise CmdLineInputError()
    group_xpath = "//group"
    group_xml = utils.get_cib_xpath(group_xpath)

    # If no groups exist, we silently return
    if group_xml == "":
        return

    element = parseString(group_xml).documentElement
    # If there is more than one group returned it's wrapped in an xpath-query
    # element
    # Ignoring mypy errors in this very old code. So far, nobody reported a bug
    # related to these lines.
    if element.tagName == "xpath-query":  # type: ignore
        elements = element.getElementsByTagName("group")  # type: ignore
    else:
        elements = [element]

    for e in elements:
        line_parts = [e.getAttribute("id") + ":"]
        line_parts.extend(
            resource.getAttribute("id")
            for resource in e.getElementsByTagName("primitive")
        )
        print(" ".join(line_parts))


def resource_status(  # noqa: PLR0912, PLR0915
    lib: Any, argv: Argv, modifiers: InputModifiers, stonith: bool = False
) -> None:
    """
    Options:
      * -f - CIB file
      * --hide-inactive - print only active resources
    """
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    del lib
    modifiers.ensure_only_supported("-f", "--hide-inactive")
    if len(argv) > 2:
        raise CmdLineInputError()

    monitor_command = ["crm_mon", "--one-shot"]
    if not modifiers.get("--hide-inactive"):
        monitor_command.append("--inactive")

    resource_or_tag_id = None
    node = None
    crm_mon_err_msg = "unable to get cluster status from crm_mon\n"
    if argv:
        for arg in argv[:]:
            if "=" not in arg:
                resource_or_tag_id = arg
                crm_mon_err_msg = f"unable to get status of '{resource_or_tag_id}' from crm_mon\n"
                monitor_command.extend(
                    [
                        "--include",
                        "none,resources",
                        "--resource",
                        resource_or_tag_id,
                    ]
                )
                argv.remove(arg)
                break
        parser = KeyValueParser(argv)
        parser.check_allowed_keys({"node"})
        node = parser.get_unique().get("node")
        if node == "":
            utils.err("missing value of 'node' option")
        if node:
            monitor_command.extend(["--node", node])

    output, retval = utils.run(monitor_command)
    if retval != 0:
        utils.err(crm_mon_err_msg + output.rstrip())
    preg = re.compile(r".*(stonith:.*)")
    resources_header = False
    in_resources = False
    has_resources = False
    no_resources_line = (
        "NO stonith devices configured"
        if stonith
        else "NO resources configured"
    )
    no_active_resources_msg = "No active resources"
    for line in output.split("\n"):
        if line in (
            "  * No active resources",  # pacemaker >= 2.0.3 with --hide-inactive
            "No active resources",  # pacemaker < 2.0.3 with --hide-inactive
        ):
            print(no_active_resources_msg)
            return
        if line in (
            "  * No resources",  # pacemaker >= 2.0.3
            "No resources",  # pacemaker < 2.0.3
        ):
            if resource_or_tag_id and not node:
                utils.err(
                    f"resource or tag id '{resource_or_tag_id}' not found"
                )
            if not node:
                print(no_resources_line)
            else:
                print(no_active_resources_msg)
            return
        if line in (
            "Full List of Resources:",  # pacemaker >= 2.0.3
            "Active Resources:",  # pacemaker >= 2.0.3 with  --hide-inactive
        ):
            in_resources = True
            continue
        if line in (
            "Full list of resources:",  # pacemaker < 2.0.3
            "Active resources:",  # pacemaker < 2.0.3 with --hide-inactive
        ):
            resources_header = True
            continue
        if line == "":
            if resources_header:
                resources_header = False
                in_resources = True
            elif in_resources:
                if not has_resources:
                    print(no_resources_line)
                return
            continue
        if in_resources:
            if resource_or_tag_id:
                has_resources = True
                print(line)
                continue
            if (
                not preg.match(line)
                and not stonith
                or preg.match(line)
                and stonith
            ):
                has_resources = True
                print(line)


def resource_disable_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --brief - show brief output of --simulate
      * --safe - only disable if no other resource gets stopped or demoted
      * --simulate - do not push the CIB, print its effects
      * --no-strict - allow disable if other resource is affected
      * --wait
    """
    if not argv:
        raise CmdLineInputError("You must specify resource(s) to disable")
    check_is_not_stonith(lib, argv, "pcs stonith disable")
    resource_disable_common(lib, argv, modifiers)


def resource_disable_common(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Commandline options:
      * -f - CIB file
      * --force - allow to disable the last stonith resource in the cluster
      * --brief - show brief output of --simulate
      * --safe - only disable if no other resource gets stopped or demoted
      * --simulate - do not push the CIB, print its effects
      * --no-strict - allow disable if other resource is affected
      * --wait
    """
    modifiers.ensure_only_supported(
        "-f",
        "--force",
        "--brief",
        "--safe",
        "--simulate",
        "--no-strict",
        "--wait",
    )
    modifiers.ensure_not_mutually_exclusive("-f", "--simulate", "--wait")
    modifiers.ensure_not_incompatible("--simulate", {"-f", "--safe", "--wait"})
    modifiers.ensure_not_incompatible("--safe", {"-f", "--simulate"})
    modifiers.ensure_not_incompatible("--no-strict", {"-f"})

    if not argv:
        raise CmdLineInputError("You must specify resource(s) to disable")

    if modifiers.get("--simulate"):
        result = lib.resource.disable_simulate(
            argv, not modifiers.get("--no-strict")
        )
        if modifiers.get("--brief"):
            # if the result is empty, printing it would produce a new line,
            # which is not wanted
            if result["other_affected_resource_list"]:
                print("\n".join(result["other_affected_resource_list"]))
            return
        print(result["plaintext_simulated_status"])
        return
    if modifiers.get("--safe") or modifiers.get("--no-strict"):
        if modifiers.get("--brief"):
            # Brief mode skips simulation output by setting the report processor
            # to ignore info reports which contain crm_simulate output and
            # resource status in this command
            lib.env.report_processor.suppress_reports_of_severity(
                [reports.ReportItemSeverity.INFO]
            )
        lib.resource.disable_safe(
            argv,
            not modifiers.get("--no-strict"),
            modifiers.get("--wait"),
        )
        return
    if modifiers.get("--brief"):
        raise CmdLineInputError(
            "'--brief' cannot be used without '--simulate' or '--safe'"
        )
    force_flags = set()
    if modifiers.get("--force"):
        force_flags.add(reports.codes.FORCE)
    lib.resource.disable(argv, modifiers.get("--wait"), force_flags)


def resource_safe_disable_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --brief - show brief output of --simulate
      * --force - skip checks for safe resource disable
      * --no-strict - allow disable if other resource is affected
      * --simulate - do not push the CIB, print its effects
      * --wait
    """
    modifiers.ensure_only_supported(
        "--brief", "--force", "--no-strict", "--simulate", "--wait"
    )
    modifiers.ensure_not_incompatible("--force", {"--no-strict", "--simulate"})
    custom_options = {}
    if modifiers.get("--force"):
        warn(
            "option '--force' is specified therefore checks for disabling "
            "resource safely will be skipped"
        )
    elif not modifiers.get("--simulate"):
        custom_options["--safe"] = True
    resource_disable_cmd(
        lib,
        argv,
        modifiers.get_subset(
            "--wait", "--no-strict", "--simulate", "--brief", **custom_options
        ),
    )


def resource_enable_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --wait
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--wait", "-f")
    if not argv:
        raise CmdLineInputError("You must specify resource(s) to enable")
    resources = argv
    check_is_not_stonith(lib, resources, "pcs stonith enable")
    lib.resource.enable(resources, modifiers.get("--wait"))


def resource_restart_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --wait
    """
    modifiers.ensure_only_supported("--wait")

    if not argv:
        raise CmdLineInputError(
            "You must specify a resource to restart",
            show_both_usage_and_message=True,
        )
    resource = argv.pop(0)
    node = argv.pop(0) if argv else None
    if argv:
        raise CmdLineInputError()

    timeout = (
        modifiers.get("--wait") if modifiers.is_specified("--wait") else None
    )

    lib.resource.restart(resource, node, timeout)

    print_to_stderr(f"{resource} successfully restarted")


def resource_force_action(  # noqa: PLR0912
    lib: Any, argv: Argv, modifiers: InputModifiers, action: str
) -> None:
    """
    Options:
      * --force
      * --full - more verbose output
    """
    # pylint: disable=too-many-branches
    modifiers.ensure_only_supported("--force", "--full")
    action_command = {
        "debug-start": "--force-start",
        "debug-stop": "--force-stop",
        "debug-promote": "--force-promote",
        "debug-demote": "--force-demote",
        "debug-monitor": "--force-check",
    }

    if action not in action_command:
        raise CmdLineInputError()
    if not argv:
        utils.err("You must specify a resource to {0}".format(action))
    if len(argv) != 1:
        raise CmdLineInputError()

    resource = argv[0]
    check_is_not_stonith(lib, [resource])
    dom = utils.get_cib_dom()

    if not (
        utils.dom_get_any_resource(dom, resource)
        or utils.dom_get_bundle(dom, resource)
    ):
        utils.err(
            "unable to find a resource/clone/group/bundle: {0}".format(resource)
        )
    bundle_el = utils.dom_get_bundle(dom, resource)
    if bundle_el:
        bundle_resource = utils.dom_get_resource_bundle(bundle_el)
        if bundle_resource:
            utils.err(
                "unable to {0} a bundle, try the bundle's resource: {1}".format(
                    action, bundle_resource.getAttribute("id")
                )
            )
        else:
            utils.err("unable to {0} a bundle".format(action))
    if utils.dom_get_group(dom, resource):
        group_resources = utils.get_group_children(resource)
        utils.err(
            (
                "unable to {0} a group, try one of the group's resource(s) ({1})"
            ).format(action, ",".join(group_resources))
        )
    if utils.dom_get_clone(dom, resource) or utils.dom_get_master(
        dom, resource
    ):
        clone_resource = utils.dom_get_clone_ms_resource(dom, resource)
        utils.err(
            "unable to {0} a clone, try the clone's resource: {1}".format(
                action, clone_resource.getAttribute("id")
            )
        )

    args = ["crm_resource", "-r", resource, action_command[action]]
    if modifiers.get("--full"):
        # set --verbose twice to get a reasonable amount of debug messages
        args.extend(["--verbose"] * 2)
    if modifiers.get("--force"):
        args.append("--force")
    output, retval = utils.run(args)

    if "doesn't support group resources" in output:
        utils.err("groups are not supported")
        sys.exit(retval)
    if "doesn't support stonith resources" in output:
        utils.err("stonith devices are not supported")
        sys.exit(retval)

    print(output.rstrip())
    sys.exit(retval)


def resource_manage_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --monitor - enable monitor operation of specified resources
    """
    modifiers.ensure_only_supported("-f", "--monitor")
    if not argv:
        raise CmdLineInputError("You must specify resource(s) to manage")
    resources = argv
    check_is_not_stonith(lib, resources)
    lib.resource.manage(resources, with_monitor=modifiers.get("--monitor"))


def resource_unmanage_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --monitor - bisable monitor operation of specified resources
    """
    modifiers.ensure_only_supported("-f", "--monitor")
    if not argv:
        raise CmdLineInputError("You must specify resource(s) to unmanage")
    resources = argv
    check_is_not_stonith(lib, resources)
    lib.resource.unmanage(resources, with_monitor=modifiers.get("--monitor"))


def resource_failcount_show(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --full
      * -f - CIB file
    """
    # pylint: disable=too-many-locals
    modifiers.ensure_only_supported("-f", "--full")

    resource = argv.pop(0) if argv and "=" not in argv[0] else None
    parser = KeyValueParser(argv)
    parser.check_allowed_keys({"node", "operation", "interval"})
    parsed_options = parser.get_unique()

    node = parsed_options.get("node")
    operation = parsed_options.get("operation")
    interval = parsed_options.get("interval")
    result_lines = []
    failures_data = lib.resource.get_failcounts(
        resource=resource, node=node, operation=operation, interval=interval
    )

    if not failures_data:
        result_lines.append(
            __headline_resource_failures(
                True, resource, node, operation, interval
            )
        )
        print("\n".join(result_lines))
        return

    resource_list = sorted({fail["resource"] for fail in failures_data})
    for current_resource in resource_list:
        result_lines.append(
            __headline_resource_failures(
                False, current_resource, node, operation, interval
            )
        )
        resource_failures = [
            fail
            for fail in failures_data
            if fail["resource"] == current_resource
        ]
        node_list = sorted({fail["node"] for fail in resource_failures})
        for current_node in node_list:
            node_failures = [
                fail
                for fail in resource_failures
                if fail["node"] == current_node
            ]
            if modifiers.get("--full"):
                result_lines.append(f"  {current_node}:")
                operation_list = sorted(
                    {fail["operation"] for fail in node_failures}
                )
                for current_operation in operation_list:
                    operation_failures = [
                        fail
                        for fail in node_failures
                        if fail["operation"] == current_operation
                    ]
                    interval_list = sorted(
                        {fail["interval"] for fail in operation_failures},
                        # pacemaker's definition of infinity
                        key=lambda x: 1000000 if x == "INFINITY" else x,
                    )
                    for current_interval in interval_list:
                        interval_failures = [
                            fail
                            for fail in operation_failures
                            if fail["interval"] == current_interval
                        ]
                        failcount, dummy_last_failure = __aggregate_failures(
                            interval_failures
                        )
                        result_lines.append(
                            f"    {current_operation} {current_interval}ms: {failcount}"
                        )
            else:
                failcount, dummy_last_failure = __aggregate_failures(
                    node_failures
                )
                result_lines.append(f"  {current_node}: {failcount}")
    print("\n".join(result_lines))


def __aggregate_failures(failure_list):
    """
    Commandline options: no options
    """
    last_failure = 0
    fail_count = 0
    for failure in failure_list:
        # infinity is a maximal value and cannot be increased
        if fail_count != "INFINITY":
            if failure["fail_count"] == "INFINITY":
                fail_count = failure["fail_count"]
            else:
                fail_count += failure["fail_count"]
        last_failure = max(last_failure, failure["last_failure"])
    return fail_count, last_failure


def __headline_resource_failures(empty, resource, node, operation, interval):
    """
    Commandline options: no options
    """
    headline_parts = []
    if empty:
        headline_parts.append("No failcounts")
    else:
        headline_parts.append("Failcounts")
    if operation:
        headline_parts.append("for operation '{operation}'")
        if interval:
            headline_parts.append("with interval '{interval}'")
    if resource:
        headline_parts.append("of" if operation else "for")
        headline_parts.append("resource '{resource}'")
    if node:
        headline_parts.append("on node '{node}'")
    return " ".join(headline_parts).format(
        node=node, resource=resource, operation=operation, interval=interval
    )


def operation_to_string(op_el):
    """
    Commandline options: no options
    """
    parts = []
    parts.append(op_el.getAttribute("name"))
    for name, value in sorted(op_el.attributes.items()):
        if name in ["id", "name"]:
            continue
        parts.append(name + "=" + value)
    parts.extend(
        f"{nvpair.getAttribute('name')}={nvpair.getAttribute('value')}"
        for nvpair in op_el.getElementsByTagName("nvpair")
    )
    parts.append("(" + op_el.getAttribute("id") + ")")
    return " ".join(parts)


def resource_cleanup(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    del lib
    modifiers.ensure_only_supported("--strict")
    resource = argv.pop(0) if argv and "=" not in argv[0] else None
    parser = KeyValueParser(argv)
    parser.check_allowed_keys({"node", "operation", "interval"})
    parsed_options = parser.get_unique()

    print_to_stderr(
        lib_pacemaker.resource_cleanup(
            utils.cmd_runner(),
            resource=resource,
            node=parsed_options.get("node"),
            operation=parsed_options.get("operation"),
            interval=parsed_options.get("interval"),
            strict=bool(modifiers.get("--strict")),
        )
    )


def resource_refresh(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - do refresh even though it may be time consuming
    """
    del lib
    modifiers.ensure_only_supported(
        "--force",
        "--strict",
        hint_syntax_changed=(
            "0.11" if modifiers.is_specified("--full") else None
        ),
    )
    resource = argv.pop(0) if argv and "=" not in argv[0] else None
    parser = KeyValueParser(argv)
    parser.check_allowed_keys({"node"})
    parsed_options = parser.get_unique()
    print_to_stderr(
        lib_pacemaker.resource_refresh(
            utils.cmd_runner(),
            resource=resource,
            node=parsed_options.get("node"),
            strict=bool(modifiers.get("--strict")),
            force=bool(modifiers.get("--force")),
        )
    )


def resource_relocate_show_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    del lib
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    resource_relocate_show(utils.get_cib_dom())


def resource_relocate_dry_run_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if argv:
        check_is_not_stonith(lib, argv)
    resource_relocate_run(utils.get_cib_dom(), argv, dry=True)


def resource_relocate_run_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        check_is_not_stonith(lib, argv)
    resource_relocate_run(utils.get_cib_dom(), argv, dry=False)


def resource_relocate_clear_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("-f")
    if argv:
        raise CmdLineInputError()
    utils.replace_cib_configuration(
        resource_relocate_clear(utils.get_cib_dom())
    )


def resource_relocate_set_stickiness(cib_dom, resources=None):
    """
    Commandline options: no options
    """
    resources = [] if resources is None else resources
    cib_dom = cib_dom.cloneNode(True)  # do not change the original cib
    resources_found = set()
    updated_resources = set()
    # set stickiness=0
    for tagname in ("master", "clone", "group", "primitive"):
        for res_el in cib_dom.getElementsByTagName(tagname):
            if resources and res_el.getAttribute("id") not in resources:
                continue
            resources_found.add(res_el.getAttribute("id"))
            res_and_children = (
                [res_el]
                + res_el.getElementsByTagName("group")
                + res_el.getElementsByTagName("primitive")
            )
            updated_resources.update(
                [el.getAttribute("id") for el in res_and_children]
            )
            for res_or_child in res_and_children:
                meta_attributes = utils.dom_prepare_child_element(
                    res_or_child,
                    "meta_attributes",
                    res_or_child.getAttribute("id") + "-meta_attributes",
                )
                utils.dom_update_nv_pair(
                    meta_attributes,
                    "resource-stickiness",
                    "0",
                    meta_attributes.getAttribute("id") + "-",
                )
    # resources don't exist
    if resources:
        resources_not_found = set(resources) - resources_found
        if resources_not_found:
            for res_id in resources_not_found:
                utils.err(
                    "unable to find a resource/clone/group: {0}".format(res_id),
                    False,
                )
            sys.exit(1)
    return cib_dom, updated_resources


def resource_relocate_get_locations(cib_dom, resources=None):
    """
    Commandline options:
      * --force - allow constraint on any resource, may not have any effective
        as an invalid constraint is ignored anyway
    """
    resources = [] if resources is None else resources
    updated_cib, updated_resources = resource_relocate_set_stickiness(
        cib_dom, resources
    )
    dummy_simout, transitions, new_cib = utils.simulate_cib(updated_cib)
    operation_list = utils.get_operations_from_transitions(transitions)
    locations = utils.get_resources_location_from_operations(
        new_cib, operation_list
    )
    # filter out non-requested resources
    if not resources:
        return list(locations.values())
    return [
        val
        for val in locations.values()
        if val["id"] in updated_resources
        or val["id_for_constraint"] in updated_resources
    ]


def resource_relocate_show(cib_dom):
    """
    Commandline options: no options
    """
    updated_cib, dummy_updated_resources = resource_relocate_set_stickiness(
        cib_dom
    )
    simout, dummy_transitions, dummy_new_cib = utils.simulate_cib(updated_cib)
    in_status = False
    in_status_resources = False
    in_transitions = False
    for line in simout.split("\n"):
        if line.strip() == "Current cluster status:":
            in_status = True
            in_status_resources = False
            in_transitions = False
        elif line.strip() == "Transition Summary:":
            in_status = False
            in_status_resources = False
            in_transitions = True
            print()
        elif line.strip() == "":
            if in_status:
                in_status = False
                in_status_resources = True
                in_transitions = False
            else:
                in_status = False
                in_status_resources = False
                in_transitions = False
        if in_status or in_status_resources or in_transitions:
            print(line)


def resource_relocate_location_to_str(location):
    """
    Commandline options: no options
    """
    message = (
        "Creating location constraint: {res} prefers {node}=INFINITY{role}"
    )
    if "start_on_node" in location:
        return message.format(
            res=location["id_for_constraint"],
            node=location["start_on_node"],
            role="",
        )
    if "promote_on_node" in location:
        return message.format(
            res=location["id_for_constraint"],
            node=location["promote_on_node"],
            role=f" role={const.PCMK_ROLE_PROMOTED}",
        )
    return ""


def resource_relocate_run(cib_dom, resources=None, dry=True):  # noqa: PLR0912
    """
    Commandline options:
      * -f - CIB file, explicitly forbids -f if dry is False
      * --force - allow constraint on any resource, may not have any effective
        as an invalid copnstraint is ignored anyway
    """
    # pylint: disable=too-many-branches
    resources = [] if resources is None else resources
    was_error = False
    anything_changed = False
    if not dry and utils.usefile:
        utils.err("This command cannot be used with -f")

    # create constraints
    cib_dom, constraint_el = constraint.getCurrentConstraints(cib_dom)
    for location in resource_relocate_get_locations(cib_dom, resources):
        if not ("start_on_node" in location or "promote_on_node" in location):
            continue
        anything_changed = True
        print_to_stderr(resource_relocate_location_to_str(location))
        constraint_id = utils.find_unique_id(
            cib_dom,
            RESOURCE_RELOCATE_CONSTRAINT_PREFIX + location["id_for_constraint"],
        )
        new_constraint = cib_dom.createElement("rsc_location")
        new_constraint.setAttribute("id", constraint_id)
        new_constraint.setAttribute("rsc", location["id_for_constraint"])
        new_constraint.setAttribute("score", "INFINITY")
        if "promote_on_node" in location:
            new_constraint.setAttribute("node", location["promote_on_node"])
            new_constraint.setAttribute(
                "role",
                pacemaker.role.get_value_for_cib(
                    const.PCMK_ROLE_PROMOTED,
                    utils.isCibVersionSatisfied(
                        cib_dom, const.PCMK_NEW_ROLES_CIB_VERSION
                    ),
                ),
            )
        elif "start_on_node" in location:
            new_constraint.setAttribute("node", location["start_on_node"])
        constraint_el.appendChild(new_constraint)
    if not anything_changed:
        return
    if not dry:
        utils.replace_cib_configuration(cib_dom)

    # wait for resources to move
    print_to_stderr("\nWaiting for resources to move...\n")
    if not dry:
        output, retval = utils.run(["crm_resource", "--wait"])
        if retval != 0:
            was_error = True
            if retval == PACEMAKER_WAIT_TIMEOUT_STATUS:
                utils.err("waiting timeout", False)
            else:
                utils.err(output, False)

    # remove constraints
    resource_relocate_clear(cib_dom)
    if not dry:
        utils.replace_cib_configuration(cib_dom)

    if was_error:
        sys.exit(1)


def resource_relocate_clear(cib_dom):
    """
    Commandline options: no options
    """
    for constraint_el in cib_dom.getElementsByTagName("constraints"):
        for location_el in constraint_el.getElementsByTagName("rsc_location"):
            location_id = location_el.getAttribute("id")
            if location_id.startswith(RESOURCE_RELOCATE_CONSTRAINT_PREFIX):
                print_to_stderr("Removing constraint {0}".format(location_id))
                location_el.parentNode.removeChild(location_el)
    return cib_dom


def set_resource_utilization(resource_id: str, argv: Argv) -> None:
    """
    Commandline options:
      * -f - CIB file
    """
    cib = utils.get_cib_dom()
    resource_el = utils.dom_get_resource(cib, resource_id)
    if resource_el is None:
        utils.err("Unable to find a resource: {0}".format(resource_id))
    utils.dom_update_utilization(resource_el, KeyValueParser(argv).get_unique())
    utils.replace_cib_configuration(cib)


def print_resource_utilization(resource_id: str) -> None:
    """
    Commandline options:
      * -f - CIB file
    """
    cib = utils.get_cib_dom()
    resource_el = utils.dom_get_resource(cib, resource_id)
    if resource_el is None:
        utils.err("Unable to find a resource: {0}".format(resource_id))
    utilization = utils.get_utilization_str(resource_el)

    print("Resource Utilization:")
    print(" {0}: {1}".format(resource_id, utilization))


def print_resources_utilization() -> None:
    """
    Commandline options:
      * -f - CIB file
    """
    cib = utils.get_cib_dom()
    utilization = {}
    for resource_el in cib.getElementsByTagName("primitive"):
        utilization_str = utils.get_utilization_str(resource_el)
        if utilization_str:
            utilization[resource_el.getAttribute("id")] = utilization_str

    print("Resource Utilization:")
    for resource in sorted(utilization):
        print(" {0}: {1}".format(resource, utilization[resource]))


# deprecated
# This is used only by pcsd, will be removed in new architecture
def get_resource_agent_info(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        utils.err("One parameter expected")
    print(json.dumps(lib.resource_agent.describe_agent(argv[0])))


def resource_bundle_create_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allow unknown options
      * --disabled - create as a stopped bundle
      * --wait
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "--disabled", "--wait", "-f")
    if not argv:
        raise CmdLineInputError()

    bundle_id = argv[0]
    parts = parse_bundle_create_options(argv[1:])
    lib.resource.bundle_create(
        bundle_id,
        parts.container_type,
        container_options=parts.container,
        network_options=parts.network,
        port_map=parts.port_map,
        storage_map=parts.storage_map,
        meta_attributes=parts.meta_attrs,
        force_options=modifiers.get("--force"),
        ensure_disabled=modifiers.get("--disabled"),
        wait=modifiers.get("--wait"),
    )


def resource_bundle_reset_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allow unknown options
      * --disabled - create as a stopped bundle
      * --wait
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "--disabled", "--wait", "-f")
    if not argv:
        raise CmdLineInputError()

    bundle_id = argv[0]
    parts = parse_bundle_reset_options(argv[1:])
    lib.resource.bundle_reset(
        bundle_id,
        container_options=parts.container,
        network_options=parts.network,
        port_map=parts.port_map,
        storage_map=parts.storage_map,
        meta_attributes=parts.meta_attrs,
        force_options=modifiers.get("--force"),
        ensure_disabled=modifiers.get("--disabled"),
        wait=modifiers.get("--wait"),
    )


def resource_bundle_update_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allow unknown options
      * --wait
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "--wait", "-f")
    if not argv:
        raise CmdLineInputError()

    bundle_id = argv[0]
    parts = parse_bundle_update_options(argv[1:])
    lib.resource.bundle_update(
        bundle_id,
        container_options=parts.container,
        network_options=parts.network,
        port_map_add=parts.port_map_add,
        port_map_remove=parts.port_map_remove,
        storage_map_add=parts.storage_map_add,
        storage_map_remove=parts.storage_map_remove,
        meta_attributes=parts.meta_attrs,
        force_options=modifiers.get("--force"),
        wait=modifiers.get("--wait"),
    )



from typing import (
    Any,
    cast,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
)
from pcs.cli.constraint import command
from pcs.cli.constraint.output import print_config
from pcs.cli.reports.preprocessor import (
    get_duplicate_constraint_exists_preprocessor,
)
from pcs.common.pacemaker.constraint import CibConstraintsDto


def create_with_set(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    create order constraint with resource set
    object lib exposes library
    list argv see usage for "constraint colocation set"
    dict like object modifiers can contain
        "force" allows resource in clone/master and constraint duplicity

    Options:
      * --force - allow resource inside clone (or master), allow duplicate
        element
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    lib.env.report_processor.set_report_item_preprocessor(
        get_duplicate_constraint_exists_preprocessor(lib)
    )
    command.create_with_set(
        lib.constraint_order.create_with_set, argv, modifiers
    )


def config_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    modifiers.ensure_only_supported("-f", "--output-format", "--full")
    if argv:
        raise CmdLineInputError()

    constraints_dto = cast(
        CibConstraintsDto,
        lib.constraint.get_config(evaluate_rules=True),
    )

    print_config(
        CibConstraintsDto(
            order=constraints_dto.order,
            order_set=constraints_dto.order_set,
        ),
        modifiers,
    )



from collections import defaultdict
from typing import (
    Any,
    Optional,
    Sequence,
    Union,
)

from pcs.cli.reports.output import error
from pcs.common import const, reports
from pcs.common.status_dto import (
    AnyResourceStatusDto,
    BundleStatusDto,
    CloneStatusDto,
    GroupStatusDto,
    PrimitiveStatusDto,
)
from pcs.common.str_tools import (
    format_list,
    format_optional,
    format_plural,
)

RESOURCE_NOT_RUNNING = "Resource '{resource_id}' is not running on any nodes"


def check_is_not_stonith(
    lib: Any, resource_id_list: list[str], cmd_to_use: Optional[str] = None
) -> None:
    if lib.resource.is_any_stonith(resource_id_list):
        raise error(
            reports.messages.CommandArgumentTypeMismatch(
                "stonith resources"
            ).message
            + format_optional(cmd_to_use, " Please use '{}' instead.")
        )


def _get_primitive_instance_list_dto(
    resource_dto: Union[AnyResourceStatusDto],
) -> Sequence[PrimitiveStatusDto]:
    """
    Return a list of primitive instances from any resource status DTO.
    """
    if isinstance(resource_dto, GroupStatusDto):
        return resource_dto.members
    if isinstance(resource_dto, CloneStatusDto):
        if isinstance(resource_dto.instances[0], GroupStatusDto):
            instance_list: list[PrimitiveStatusDto] = []
            for group_dto in resource_dto.instances:
                # There is a check for the type in the if above
                instance_list.extend(group_dto.members)  # type: ignore
            return instance_list
        # There is a check for the type in the if above
        return resource_dto.instances  # type: ignore
    if isinstance(resource_dto, BundleStatusDto):
        return [replica_dto.container for replica_dto in resource_dto.replicas]
    return [resource_dto]


def get_resource_status_msg(lib: Any, resource_id: str) -> str:
    """
    Get where resources are running, typically used after waiting for cluster
    is finished. Returns a text string similar to utils.resource_running_on.
    Examples:
        Resource 'r1' is not running on any nodes
        Resource 'r2-clone' is promoted on node 'n1'; unpromoted on nodes 'n2',
        'n3'
    """
    resource_dto = None
    for resource in lib.status.resources_status().resources:
        if resource.resource_id == resource_id:
            resource_dto = resource
    if resource_dto is None:
        # Resource is configured but Pacemaker is ignoring it
        return RESOURCE_NOT_RUNNING.format(resource_id=resource_id)

    # Using set in case more instances are running on the same node
    role_and_location: dict[const.PcmkStatusRoleType, set[str]] = defaultdict(
        set
    )
    instance_list_dto = _get_primitive_instance_list_dto(resource_dto)
    for instance_dto in instance_list_dto:
        role_and_location[instance_dto.role].update(instance_dto.node_names)

    # The old function (utils.resource_running_on) this is replicating only
    # supported states Started, Promoted and Unpromoted. Therefore we have to
    # filter out only these states since others were ignored previously.
    role_and_location = {
        role: loc
        for role, loc in role_and_location.items()
        if role
        in [
            const.PCMK_STATUS_ROLE_STARTED,
            const.PCMK_STATUS_ROLE_PROMOTED,
            const.PCMK_STATUS_ROLE_UNPROMOTED,
        ]
    }

    if not role_and_location:
        return RESOURCE_NOT_RUNNING.format(resource_id=resource_id)

    state_parts = []
    for state_name, node_list in role_and_location.items():
        state_parts.append(
            "{state_name} on {node_pl} {node_list}".format(
                state_name=(
                    state_name.lower()
                    if state_name != const.PCMK_STATUS_ROLE_STARTED
                    else "running"
                ),
                node_pl=format_plural(depends_on=node_list, singular="node"),
                node_list=format_list(node_list),
            )
        )
    state_info = "; ".join(state_parts)
    return f"Resource '{resource_id}' is {state_info}"



from unittest import TestCase

from pcs.common.pacemaker.constraint import (
    CibConstraintsDto,
    get_all_constraints_ids,
    get_all_location_constraints_ids,
    get_all_location_rules_ids,
)

from pcs_test.tools.constraints_dto import get_all_constraints
from pcs_test.tools.custom_mock import RuleInEffectEvalMock


class GetAllConstraintsIds(TestCase):
    def test_constraints_defined(self):
        self.assertEqual(
            get_all_constraints_ids(
                get_all_constraints(RuleInEffectEvalMock({}))
            ),
            {
                "colocation-G1-clone-R6-clone--100",
                "colocation_set_R7G2B2",
                "loc_constr_with_expired_rule",
                "loc_constr_with_not_expired_rule",
                "location-G2-localhost-INFINITY",
                "location-R-localhost-INFINITY",
                "location-R7-another-one--INFINITY",
                "location-R7-localhost-INFINITY",
                "location-R7-non-existing-node--10000",
                "order-G2-B2-Optional",
                "order-R7-G2-mandatory",
                "order_set_B2R6-cloneSe",
                "ticket-custom-ticket1-G1-clone-Promoted",
                "ticket_set_R7B2G2",
            },
        )

    def test_constraints_not_defined(self):
        self.assertEqual(get_all_constraints_ids(CibConstraintsDto()), set())


class GetAllLocationRulesIds(TestCase):
    def test_location_constraints_with_rules(self):
        constraint_dto = get_all_constraints(RuleInEffectEvalMock({}))
        self.assertEqual(
            get_all_location_rules_ids(constraint_dto),
            {
                "loc_constr_with_not_expired_rule-rule-1",
                "loc_constr_with_expired_rule-rule",
                "loc_constr_with_not_expired_rule-rule",
            },
        )

    def test_location_constraints_without_rules(self):
        constraint_dto = get_all_constraints(RuleInEffectEvalMock({}))
        location_constraints_without_rules = CibConstraintsDto(
            location=[
                location_dto
                for location_dto in constraint_dto.location
                if not location_dto.attributes.rules
            ]
        )
        self.assertEqual(
            get_all_location_rules_ids(location_constraints_without_rules),
            set(),
        )


class GetAllLocationIds(TestCase):
    def test_location_constraints_defined(self):
        constraint_dto = get_all_constraints(RuleInEffectEvalMock({}))
        self.assertEqual(
            get_all_location_constraints_ids(constraint_dto),
            {
                "location-G2-localhost-INFINITY",
                "location-R-localhost-INFINITY",
                "location-R7-another-one--INFINITY",
                "location-R7-localhost-INFINITY",
                "location-R7-non-existing-node--10000",
                "loc_constr_with_expired_rule",
                "loc_constr_with_not_expired_rule",
            },
        )

    def test_location_constraints_not_defined(self):
        constraint_dto = get_all_constraints(RuleInEffectEvalMock({}))
        no_location_constraints_dto = CibConstraintsDto(
            colocation=constraint_dto.colocation,
            colocation_set=constraint_dto.colocation_set,
            order=constraint_dto.order,
            order_set=constraint_dto.order_set,
            ticket=constraint_dto.ticket_set,
            ticket_set=constraint_dto.ticket_set,
        )
        self.assertEqual(
            get_all_location_constraints_ids(no_location_constraints_dto),
            set(),
        )



from collections import Counter
from functools import partial
from typing import (
    Final,
    Mapping,
    Optional,
    Union,
)

from pcs.cli.common.errors import (
    SEE_MAN_CHANGES,
    CmdLineInputError,
)
from pcs.cli.reports.output import deprecation_warning
from pcs.common.const import INFINITY
from pcs.common.str_tools import (
    format_list,
    format_list_custom_last_separator,
    format_plural,
)
from pcs.common.tools import timeout_to_seconds
from pcs.common.types import (
    StringCollection,
    StringIterable,
    StringSequence,
)

# sys.argv always returns a list, we don't need StringSequence in here
Argv = list[str]
ModifierValueType = Union[None, bool, str]

_FUTURE_OPTION_STR: Final = "future"
FUTURE_OPTION: Final = f"--{_FUTURE_OPTION_STR}"
_OUTPUT_FORMAT_OPTION_STR: Final = "output-format"
OUTPUT_FORMAT_OPTION: Final = f"--{_OUTPUT_FORMAT_OPTION_STR}"
OUTPUT_FORMAT_VALUE_CMD: Final = "cmd"
OUTPUT_FORMAT_VALUE_JSON: Final = "json"
OUTPUT_FORMAT_VALUE_TEXT: Final = "text"
OUTPUT_FORMAT_VALUES: Final = frozenset(
    (
        OUTPUT_FORMAT_VALUE_CMD,
        OUTPUT_FORMAT_VALUE_JSON,
        OUTPUT_FORMAT_VALUE_TEXT,
    )
)

ARG_TYPE_DELIMITER: Final = "%"

# h = help, f = file,
# p = password (cluster auth), u = user (cluster auth),
PCS_SHORT_OPTIONS: Final = "hf:p:u:"
PCS_LONG_OPTIONS: Final = [
    "debug",
    "version",
    "help",
    "fullhelp",
    "force",
    "skip-offline",
    "autodelete",
    "simulate",
    "all",
    "full",
    "local",
    "wait",
    "config",
    "start",
    "enable",
    "disabled",
    "off",
    "request-timeout=",
    "brief",
    _FUTURE_OPTION_STR,
    # resource (safe-)disable
    "safe",
    "no-strict",
    # resource cleanup | refresh
    "strict",
    "pacemaker",
    "corosync",
    "no-default-ops",
    "defaults",
    "nodesc",
    "promoted",
    "name=",
    "group=",
    "node=",
    "from=",
    "to=",
    "after=",
    "before=",
    "corosync_conf=",
    "booth-conf=",
    "booth-key=",
    # do not stop resources in resource delete command
    "no-stop",
    "no-watchdog-validation",
    # pcs cluster setup
    "no-cluster-uuid",
    "no-keys-sync",
    # in pcs status - do not display resource status on inactive node
    "hide-inactive",
    # pcs resource (un)manage - enable or disable monitor operations
    "monitor",
    # TODO remove
    # used only in deprecated 'pcs resource|stonith show'
    "groups",
    # "pcs resource clear --expired" - only clear expired moves and bans
    "expired",
    # disable evaluating whether rules are expired
    "no-expire-check",
    # allow overwriting existing files, currently meant for / used in CLI only
    "overwrite",
    # output format of commands, e.g: json, cmd, text, ...
    f"{_OUTPUT_FORMAT_OPTION_STR}=",
    # auth token
    "token=",
    # enable agent self validation
    "agent-validation",
    # disable text output in query commands
    "quiet",
    # proceed with dangerous actions, meant for / used in CLI only
    "yes",
]


def split_list(arg_list: Argv, separator: str) -> list[Argv]:
    """
    split a list of arguments to several lists using separator as a delimiter

    arg_list -- list of command line arguments to split
    separator -- delimiter
    """
    separator_indexes = [i for i, x in enumerate(arg_list) if x == separator]
    bounds = zip(
        [0] + [i + 1 for i in separator_indexes],
        separator_indexes + [None],
        strict=False,
    )
    return [arg_list[i:j] for i, j in bounds]


def split_list_by_any_keywords(
    arg_list: Argv, keyword_label: str
) -> dict[str, Argv]:
    """
    split a list of arguments using any argument not containing = as a delimiter

    arg_list -- list of command line arguments to split
    keyword_label -- description of all keywords
    """
    groups: dict[str, Argv] = {}
    if not arg_list:
        return groups

    if "=" in arg_list[0]:
        raise CmdLineInputError(
            f"Invalid character '=' in {keyword_label} '{arg_list[0]}'"
        )

    current_keyword = arg_list[0]
    groups[current_keyword] = []
    for arg in arg_list[1:]:
        if "=" in arg:
            groups[current_keyword].append(arg)
        else:
            current_keyword = arg
            if current_keyword in groups:
                raise CmdLineInputError(
                    "{} '{}' defined multiple times".format(
                        keyword_label.capitalize(), current_keyword
                    )
                )
            groups[current_keyword] = []
    return groups


def split_option(arg: str, allow_empty_value: bool = True) -> tuple[str, str]:
    """
    Get (key, value) from a key=value commandline argument.

    Split the argument by the first = and return resulting parts. Raise
    CmdLineInputError if the argument cannot be split.

    arg -- commandline argument to split
    allow_empty_value -- if False, raise CmdLineInputError on empty value
    """
    if "=" not in arg:
        raise CmdLineInputError(f"missing value of '{arg}' option")
    if arg.startswith("="):
        raise CmdLineInputError(f"missing key in '{arg}' option")
    key, value = arg.split("=", 1)
    if not (value or allow_empty_value):
        raise CmdLineInputError(f"value of '{key}' option is empty")
    return key, value


def ensure_unique_args(cmdline_args: Argv) -> None:
    """
    Raises in case there are duplicate args
    """
    duplicities = [
        item for item, count in Counter(cmdline_args).items() if count > 1
    ]
    if duplicities:
        argument_pl = format_plural(duplicities, "argument")
        duplicities_list = format_list(duplicities)
        raise CmdLineInputError(f"duplicate {argument_pl}: {duplicities_list}")


class KeyValueParser:
    """
    Parse and check key=value options
    """

    def __init__(self, arg_list: Argv, repeatable: StringCollection = ()):
        """
        arg_list -- commandline arguments to be parsed
        repeatable -- keys that are allowed to be specified several times
        """
        self._repeatable_keys = repeatable
        self._key_value_map: dict[str, list[str]] = {}
        for arg in arg_list:
            name, value = split_option(arg)
            if name not in self._key_value_map:
                self._key_value_map[name] = [value]
            else:
                self._key_value_map[name].append(value)

    def check_allowed_keys(self, allowed_keys: StringCollection) -> None:
        """
        Check that only allowed keys were specified

        allowed_keys -- list of allowed keys
        """
        unknown_options = set(self._key_value_map.keys()) - set(allowed_keys)
        if unknown_options:
            raise CmdLineInputError(
                "Unknown option{s} '{options}'".format(
                    s=("s" if len(unknown_options) > 1 else ""),
                    options="', '".join(sorted(unknown_options)),
                )
            )

    def get_unique(self) -> dict[str, str]:
        """
        Get all non-repeatable keys and their values; raise if a key has more values
        """
        result: dict[str, str] = {}
        for key, values in self._key_value_map.items():
            if key in self._repeatable_keys:
                continue
            values_uniq = set(values)
            if len(values_uniq) > 1:
                raise CmdLineInputError(
                    f"duplicate option '{key}' with different values "
                    f"{format_list_custom_last_separator(values_uniq, ' and ')}"
                )
            result[key] = values[0]
        return result

    def get_repeatable(self) -> dict[str, list[str]]:
        """
        Get all repeatable keys and their values
        """
        return {
            key: self._key_value_map[key]
            for key in self._repeatable_keys
            if key in self._key_value_map
        }


class ArgsByKeywords:
    def __init__(self, groups: Mapping[str, list[Argv]]):
        self._groups = groups
        self._flat_cache: dict[str, Argv] = {}

    def allow_repetition_only_for(self, keyword_set: StringCollection) -> None:
        """
        Raise CmdLineInputError if a keyword has been repeated when not allowed

        keyword_set -- repetition is allowed for these keywords
        """
        for keyword, arg_groups in self._groups.items():
            if len(arg_groups) > 1 and keyword not in keyword_set:
                raise CmdLineInputError(
                    f"'{keyword}' cannot be used more than once"
                )

    def ensure_unique_keywords(self) -> None:
        """
        Raise CmdLineInputError if any keyword has been repeated
        """
        return self.allow_repetition_only_for(set())

    def is_empty(self) -> bool:
        """
        Check if any args have been specified
        """
        return not self._groups

    def has_keyword(self, keyword: str) -> bool:
        """
        Check if a keyword has been specified

        keyword -- a keyword to check
        """
        return keyword in self._groups

    def has_empty_keyword(self, keyword: str) -> bool:
        """
        Check if a keyword has been specified without any following args

        keyword -- a keyword to check
        """
        return self.has_keyword(keyword) and not self.get_args_flat(keyword)

    def get_args_flat(self, keyword: str) -> Argv:
        """
        Get arguments of a keyword in one sequence
        """
        if keyword in self._groups:
            if keyword not in self._flat_cache:
                self._flat_cache[keyword] = [
                    arg
                    for one_group in self._groups[keyword]
                    for arg in one_group
                ]
            return self._flat_cache[keyword]
        return []

    def get_args_groups(self, keyword: str) -> list[Argv]:
        """
        Get arguments of a keyword, one group for each keyword occurrence
        """
        if keyword in self._groups:
            return self._groups[keyword]
        return []


def group_by_keywords(
    arg_list: Argv,
    keyword_set: StringCollection,
    implicit_first_keyword: Optional[str] = None,
) -> ArgsByKeywords:
    """
    Separate argv into groups delimited by specified keywords

    arg_list -- commandline arguments containing keywords
    keyword_set -- all expected keywords
    implicit_first_keyword -- key for capturing args before the first keyword
    """
    args_by_keywords: dict[str, list[Argv]] = {}

    def new_keyword(keyword: str) -> None:
        if keyword not in args_by_keywords:
            args_by_keywords[keyword] = []
        args_by_keywords[keyword].append([])

    if arg_list:
        if arg_list[0] not in keyword_set:
            if not implicit_first_keyword:
                raise CmdLineInputError()
            current_keyword = implicit_first_keyword
            new_keyword(current_keyword)

        for arg in arg_list:
            if arg in keyword_set:
                current_keyword = arg
                new_keyword(current_keyword)
            else:
                args_by_keywords[current_keyword][-1].append(arg)

    return ArgsByKeywords(args_by_keywords)


def parse_typed_arg(
    arg: str, allowed_types: StringSequence, default_type: str
) -> tuple[str, str]:
    """
    Get (type, value) from a typed commandline argument.

    Split the argument by the type separator and return the type and the value.
    Raise CmdLineInputError in the argument format or type is not valid.
    string arg -- commandline argument
    Iterable allowed_types -- list of allowed argument types
    string default_type -- type to return if the argument doesn't specify a type
    """
    if ARG_TYPE_DELIMITER not in arg:
        return default_type, arg
    arg_type, arg_value = arg.split(ARG_TYPE_DELIMITER, 1)
    if not arg_type:
        return default_type, arg_value
    if arg_type not in allowed_types:
        raise CmdLineInputError(
            (
                "'{arg_type}' is not an allowed type for '{arg_full}', use "
                "{hint}"
            ).format(
                arg_type=arg_type,
                arg_full=arg,
                hint=", ".join(sorted(allowed_types)),
            )
        )
    return arg_type, arg_value


def _is_num(arg: str) -> bool:
    if arg.lower() == INFINITY.lower():
        return True
    try:
        int(arg)
        return True
    except ValueError:
        return False


def _is_float(arg: str) -> bool:
    try:
        float(arg)
        return True
    except ValueError:
        return False


def _is_negative_num(arg: str) -> bool:
    return arg.startswith("-") and (_is_num(arg[1:]) or _is_float(arg))


def is_short_option_expecting_value(arg: str) -> bool:
    return len(arg) == 2 and arg[0] == "-" and f"{arg[1]}:" in PCS_SHORT_OPTIONS


def is_long_option_expecting_value(arg: str) -> bool:
    return (
        len(arg) > 2 and arg[0:2] == "--" and f"{arg[2:]}=" in PCS_LONG_OPTIONS
    )


def is_option_expecting_value(arg: str) -> bool:
    return is_short_option_expecting_value(
        arg
    ) or is_long_option_expecting_value(arg)


# DEPRECATED
# TODO remove
# This function is called only by deprecated code for parsing argv containing
# negative numbers without -- prepending them.
def filter_out_non_option_negative_numbers(arg_list: Argv) -> tuple[Argv, Argv]:
    """
    Return arg_list without non-option negative numbers.
    Negative numbers following the option expecting value are kept.

    There is the problematic legacy:
    Argument "--" has special meaning: it can be used to signal that no more
    options will follow. This would solve the problem with negative numbers in
    a standard way: there would be no special approach to negative numbers,
    everything would be left in the hands of users.

    We cannot use "--" as it would be a backward incompatible change:
    * "pcs ... -infinity" would not work any more, users would have to switch
      to "pcs ... -- ... -infinity"
    * previously, position of some --options mattered, for example
      "--clone <clone options>", this syntax would not be possible with the "--"
      in place

    Currently used --options, which may be problematic when switching to "--":
    * --group <group name>, --before | --after <resource id>
      * pcs resource | stonith create, pcs resource group add, pcs tag update
      * They have a single argument, so they would work even with --. But the
        command may look weird:
        pcs resource create --group G --after R2 -- R3 ocf:pacemaker:Dummy
        vs. current command
        pcs resource create R3 ocf:pacemaker:Dummy --group G --after R2

    list arg_list contains command line arguments
    """
    args_without_negative_nums = []
    args_filtered_out = []
    for i, arg in enumerate(arg_list):
        prev_arg = arg_list[i - 1] if i > 0 else ""
        if not _is_negative_num(arg) or is_option_expecting_value(prev_arg):
            args_without_negative_nums.append(arg)
        else:
            args_filtered_out.append(arg)

    return args_without_negative_nums, args_filtered_out


# DEPRECATED
# TODO remove
# This function is called only by deprecated code for parsing argv containing
# negative numbers without -- prepending them.
def filter_out_options(arg_list: Argv) -> Argv:
    """
    Return arg_list without options and negative numbers

    See a comment in filter_out_non_option_negative_numbers.

    arg_list -- command line arguments
    """
    args_without_options = []
    for i, arg in enumerate(arg_list):
        prev_arg = arg_list[i - 1] if i > 0 else ""
        if not is_option_expecting_value(prev_arg) and (
            not arg.startswith("-") or arg == "-" or _is_negative_num(arg)
        ):
            args_without_options.append(arg)
    return args_without_options


def wait_to_timeout(wait: Union[bool, str, None]) -> int:
    if wait is False:
        return -1
    if wait is None:
        return 0
    timeout = timeout_to_seconds(wait)
    if timeout is None:
        raise CmdLineInputError(f"'{wait}' is not a valid interval value")
    return timeout


class InputModifiers:
    def __init__(self, options: Mapping[str, ModifierValueType]):
        self._defined_options = set(options.keys())
        self._options = dict(options)
        self._options.update(
            {
                # boolean values
                "--all": "--all" in options,
                "--agent-validation": "--agent-validation" in options,
                "--autodelete": "--autodelete" in options,
                "--brief": "--brief" in options,
                "--config": "--config" in options,
                "--corosync": "--corosync" in options,
                "--debug": "--debug" in options,
                "--defaults": "--defaults" in options,
                "--disabled": "--disabled" in options,
                "--enable": "--enable" in options,
                "--expired": "--expired" in options,
                "--force": "--force" in options,
                "--full": "--full" in options,
                "--quiet": "--quiet" in options,
                FUTURE_OPTION: FUTURE_OPTION in options,
                # TODO remove
                # used only in deprecated 'pcs resource|stonith show'
                "--groups": "--groups" in options,
                "--hide-inactive": "--hide-inactive" in options,
                "--local": "--local" in options,
                "--monitor": "--monitor" in options,
                "--no-default-ops": "--no-default-ops" in options,
                "--nodesc": "--nodesc" in options,
                "--no-expire-check": "--no-expire-check" in options,
                "--no-cluster-uuid": "--no-cluster-uuid" in options,
                "--no-keys-sync": "--no-keys-sync" in options,
                "--no-stop": "--no-stop" in options,
                "--no-strict": "--no-strict" in options,
                "--no-watchdog-validation": (
                    "--no-watchdog-validation" in options
                ),
                "--off": "--off" in options,
                "--overwrite": "--overwrite" in options,
                "--pacemaker": "--pacemaker" in options,
                "--promoted": "--promoted" in options,
                "--safe": "--safe" in options,
                "--simulate": "--simulate" in options,
                "--skip-offline": "--skip-offline" in options,
                "--start": "--start" in options,
                "--strict": "--strict" in options,
                "--yes": "--yes" in options,
                # string values
                "--after": options.get("--after", None),
                "--before": options.get("--before", None),
                "--booth-conf": options.get("--booth-conf", None),
                "--booth-key": options.get("--booth-key", None),
                "--corosync_conf": options.get("--corosync_conf", None),
                "--from": options.get("--from", None),
                # TODO remove
                # used in resource create and stonith create, deprecated in both
                "--group": options.get("--group", None),
                "--name": options.get("--name", None),
                "--node": options.get("--node", None),
                OUTPUT_FORMAT_OPTION: options.get(
                    OUTPUT_FORMAT_OPTION, OUTPUT_FORMAT_VALUE_TEXT
                ),
                "--request-timeout": options.get("--request-timeout", None),
                "--to": options.get("--to", None),
                "--token": options.get("--token", None),
                "--wait": options.get("--wait", False),
                "-f": options.get("-f", None),
                "-p": options.get("-p", None),
                "-u": options.get("-u", None),
            }
        )

    def get_subset(
        self, *options: str, **custom_options: ModifierValueType
    ) -> "InputModifiers":
        opt_dict = {
            opt: self.get(opt) for opt in options if self.is_specified(opt)
        }
        opt_dict.update(custom_options)
        return InputModifiers(opt_dict)

    def ensure_only_supported(
        self,
        *supported_options: str,
        hint_syntax_changed: Optional[str] = None,
        output_format_supported: bool = False,
    ) -> None:
        # --debug is supported in all commands
        supported_options_set = set(supported_options) | {"--debug"}
        if output_format_supported:
            supported_options_set.add(OUTPUT_FORMAT_OPTION)
        unsupported_options = self._defined_options - supported_options_set
        if unsupported_options:
            pluralize = partial(format_plural, unsupported_options)
            raise CmdLineInputError(
                "Specified {option} {option_list} {_is} not supported in this "
                "command".format(
                    option=pluralize("option"),
                    option_list=format_list(sorted(unsupported_options)),
                    _is=pluralize("is"),
                ),
                hint=(
                    "Syntax has changed from previous version. {}".format(
                        SEE_MAN_CHANGES.format(hint_syntax_changed)
                    )
                    if hint_syntax_changed
                    else None
                ),
            )

    def ensure_not_mutually_exclusive(self, *mutually_exclusive: str) -> None:
        """
        Raise CmdLineInputError if several exclusive options were specified

        mutually_exclusive -- mutually exclusive options
        """
        options_to_report = self._defined_options & set(mutually_exclusive)
        if len(options_to_report) > 1:
            raise CmdLineInputError(
                "Only one of {} can be used".format(
                    format_list(sorted(options_to_report))
                )
            )

    def ensure_not_incompatible(
        self, checked: str, incompatible: StringCollection
    ) -> None:
        """
        Raise CmdLineInputError if both the checked and an incompatible option
        were specified

        checked -- option incompatible with any of incompatible options
        incompatible -- set of options incompatible with checked
        """
        if checked not in self._defined_options:
            return
        disallowed = self._defined_options & set(incompatible)
        if disallowed:
            raise CmdLineInputError(
                "'{}' cannot be used with {}".format(
                    checked, format_list(sorted(disallowed))
                )
            )

    def ensure_dependency_satisfied(
        self, main_option: str, dependent_options: StringCollection
    ) -> None:
        """
        Raise CmdLineInputError if any of dependent_options is present and
        main_option is not present.

        main_option -- option on which dependent_options depend
        dependent_options -- none of these options can be specified if
            main_option is not specified
        """
        if main_option in self._defined_options:
            return
        disallowed = self._defined_options & set(dependent_options)
        if disallowed:
            raise CmdLineInputError(
                "{} cannot be used without '{}'".format(
                    format_list(sorted(disallowed)), main_option
                )
            )

    def is_specified(self, option: str) -> bool:
        return option in self._defined_options

    def is_specified_any(self, option_list: StringIterable) -> bool:
        return any(self.is_specified(option) for option in option_list)

    def get(
        self, option: str, default: ModifierValueType = None
    ) -> ModifierValueType:
        if option in self._defined_options:
            return self._options[option]
        if default is not None:
            return default
        if option in self._options:
            return self._options[option]
        raise AssertionError(f"Non existing default value for '{option}'")

    def get_output_format(
        self,
        supported_formats: StringCollection = OUTPUT_FORMAT_VALUES,
    ) -> str:
        output_format = self.get(OUTPUT_FORMAT_OPTION)
        if output_format in supported_formats:
            return str(output_format)
        raise CmdLineInputError(
            (
                "Unknown value '{value}' for '{option}' option. Supported "
                "{value_pl} {is_pl}: {supported}"
            ).format(
                value=output_format,
                option=OUTPUT_FORMAT_OPTION,
                value_pl=format_plural(supported_formats, "value"),
                is_pl=format_plural(supported_formats, "is"),
                supported=format_list(list(supported_formats)),
            )
        )


def get_rule_str(argv: Argv) -> Optional[str]:
    if argv:
        if len(argv) > 1:
            # deprecated after 0.11.7
            deprecation_warning(
                "Specifying a rule as multiple arguments is deprecated and "
                "might be removed in a future release, specify the rule as "
                "a single string instead"
            )
            return " ".join(argv)
        return argv[0]
    return None



from unittest import TestCase

from lxml import etree

from pcs.common import const, reports
from pcs.lib.cib.constraint.common import (
    DuplicatesChecker,
    create_constraint_with_set,
    find_constraints_of_same_type,
    is_constraint,
    validate_constrainable_elements,
)
from pcs.lib.cib.tools import IdProvider, Version

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.fixture import ReportItemFixture
from pcs_test.tools.xml import etree_to_str, str_to_etree


class IsConstraint(TestCase):
    def test_is_constraint_true(self):
        for element in (
            etree.Element("rsc_colocation"),
            etree.Element("rsc_location"),
            etree.Element("rsc_order"),
            etree.Element("rsc_ticket"),
        ):
            with self.subTest(element=element):
                self.assertTrue(is_constraint(element))

    def test_is_constraint_false(self):
        self.assertFalse(is_constraint(etree.Element("element")))


def fixture_cib():
    res_set1 = """
        <resource_set>
            <resource_ref id="R1"/> <resource_ref id="R2"/>
        </resource_set>
    """
    res_set2 = """
        <resource_set>
            <resource_ref id="R2"/> <resource_ref id="R3"/>
        </resource_set>
    """
    res_set3 = """
        <resource_set>
            <resource_ref id="R1"/> <resource_ref id="R3"/>
        </resource_set>
    """
    return str_to_etree(
        f"""
            <constraints>
                <rsc_location id="LP1" rsc="R1" node="node1" score="1" />
                <rsc_location id="LP2" rsc="R2" node="node2" score="2" />
                <rsc_location id="LP3" rsc="R3" node="node3" score="3" />
                <rsc_location id="LS1" node="node1" score="1">{res_set1}</rsc_location>
                <rsc_location id="LS2" node="node2" score="2">{res_set2}</rsc_location>
                <rsc_location id="LS3" node="node3" score="3">{res_set3}</rsc_location>
                <rsc_colocation id="CP1" rsc="R1" with-rsc="R2" score="1" />
                <rsc_colocation id="CP2" rsc="R1" with-rsc="R3" score="1" />
                <rsc_colocation id="CP3" rsc="R2" with-rsc="R3" score="1" />
                <rsc_colocation id="CS1" score="1">{res_set1}</rsc_colocation>
                <rsc_colocation id="CS2" score="2">{res_set2}</rsc_colocation>
                <rsc_colocation id="CS3" score="3">{res_set3}</rsc_colocation>
                <rsc_order id="OP1" first="R1" then="R2" />
                <rsc_order id="OP2" first="R2" then="R3" />
                <rsc_order id="OP3" first="R1" then="R3" />
                <rsc_order id="OS1">{res_set1}</rsc_order>
                <rsc_order id="OS2">{res_set2}</rsc_order>
                <rsc_order id="OS3">{res_set3}</rsc_order>
                <rsc_ticket id="TP1" rsc="R1" ticket="T1" />
                <rsc_ticket id="TP2" rsc="R2" ticket="T2" />
                <rsc_ticket id="TP3" rsc="R3" ticket="T3" />
                <rsc_ticket id="TS1" ticket="T1">{res_set1}</rsc_ticket>
                <rsc_ticket id="TS2" ticket="T2">{res_set2}</rsc_ticket>
                <rsc_ticket id="TS3" ticket="T3">{res_set3}</rsc_ticket>
            </constraints>
        """
    )


class FindConstraintsOfSameType(TestCase):
    def test_found(self):
        cib = fixture_cib()
        for type_id in ("LP", "LS", "CP", "CS", "OP", "OS", "TP", "TS"):
            with self.subTest(constraint_type=type_id):
                element = cib.xpath(".//*[@id=$id]", id=f"{type_id}1")[0]
                self.assertEqual(
                    [
                        el.attrib["id"]
                        for el in find_constraints_of_same_type(cib, element)
                    ],
                    [f"{type_id}2", f"{type_id}3"],
                )

    def test_not_found(self):
        cib = fixture_cib()
        for element in cib.xpath("./*[not(contains(@id, '1'))]"):
            element.getparent().remove(element)

        for type_id in ("LP", "LS", "CP", "CS", "OP", "OS", "TP", "TS"):
            with self.subTest(constraint_type=type_id):
                element = cib.xpath(".//*[@id=$id]", id=f"{type_id}1")[0]
                self.assertEqual(
                    [
                        el.attrib["id"]
                        for el in find_constraints_of_same_type(cib, element)
                    ],
                    [],
                )


class DuplicatesCheckerTestBase(TestCase):
    def assert_success(self, cib, checker, duplicates):
        for id_to_check, id_results in duplicates.items():
            for forced in (False, True):
                with self.subTest(id_to_check=id_to_check, forced=forced):
                    real_reports = checker.check(
                        cib,
                        cib.xpath(".//*[@id=$id]", id=f"{id_to_check}")[0],
                        force_flags=([reports.codes.FORCE] if forced else []),
                    )
                    expected_reports = []
                    if id_results:
                        if forced:
                            expected_reports = [
                                fixture.warn(
                                    reports.codes.DUPLICATE_CONSTRAINTS_EXIST,
                                    constraint_ids=id_results,
                                )
                            ]
                        else:
                            expected_reports = [
                                fixture.error(
                                    reports.codes.DUPLICATE_CONSTRAINTS_EXIST,
                                    force_code=reports.codes.FORCE,
                                    constraint_ids=id_results,
                                )
                            ]
                    assert_report_item_list_equal(
                        real_reports, expected_reports
                    )


class DuplicatesCheckerTest(DuplicatesCheckerTestBase):
    class MockChecker(DuplicatesChecker):
        def _are_duplicate(self, constraint_to_check, constraint_el):
            return (
                int(constraint_to_check.attrib["id"][-1]) % 2
                == int(constraint_el.attrib["id"][-1]) % 2
            )

    def test_success(self):
        cib = fixture_cib()
        duplicates = {
            "LP1": ["LP3"],
            "LP2": [],
            "LP3": ["LP1"],
            "LS1": ["LS3"],
            "LS2": [],
            "LS3": ["LS1"],
            "CP1": ["CP3"],
            "CP2": [],
            "CP3": ["CP1"],
            "CS1": ["CS3"],
            "CS2": [],
            "CS3": ["CS1"],
            "OP1": ["OP3"],
            "OP2": [],
            "OP3": ["OP1"],
            "OS1": ["OS3"],
            "OS2": [],
            "OS3": ["OS1"],
            "TP1": ["TP3"],
            "TP2": [],
            "TP3": ["TP1"],
            "TS1": ["TS3"],
            "TS2": [],
            "TS3": ["TS1"],
        }
        checker = self.MockChecker()
        self.assert_success(cib, checker, duplicates)


class ValidateConstrainableElement(TestCase):
    _cib = str_to_etree(
        """
            <resources>
              <bundle id="B">
                <docker image="pcs:test" />
                <primitive id="B_R" class="ocf" type="Dummy" provider="pacemaker" />
              </bundle>
              <group id="G">
                <primitive id="G_R1" class="ocf" type="Dummy" provider="pacemaker" />
                <primitive id="G_R2" class="ocf" type="Dummy" provider="pacemaker" />
              </group>
              <clone id="C1">
                <group id="C1_G">
                  <primitive id="C1_G_R1" class="ocf" type="Stateful" provider="pacemaker" />
                  <primitive id="C1_G_R2" class="ocf" type="Stateful" provider="pacemaker" />
                </group>
              </clone>
              <clone id="C2">
                <primitive id="C2_R" class="ocf" type="Dummy" provider="pacemaker" />
              </clone>
              <primitive id="R" class="ocf" type="Dummy" provider="pacemaker" />
            </resources>
        """
    )

    def test_no_report(self):
        id_map = {
            "R": "primitive",
            "G_R1": "primitive in a group",
            "G": "group",
            "C1": "clone",
            "B": "bundle",
        }
        for res_id, res_desc in id_map.items():
            with self.subTest(
                resource_id=res_id, resource_description=res_desc
            ):
                assert_report_item_list_equal(
                    validate_constrainable_elements(
                        self._cib.xpath("//*[@id=$id]", id=res_id)
                    ),
                    [],
                )

    def _test_report(self, in_multiinstance_allowed, severity, force_code):
        id_map = {
            "B_R": "primitive in a bundle",
            "C1_G": "group in a clone",
            "C1_G_R1": "primitive in a group in a clone",
            "C2_R": "primitive in a clone",
        }
        for res_id, res_desc in id_map.items():
            with self.subTest(
                resource_id=res_id, resource_description=res_desc
            ):
                assert_report_item_list_equal(
                    validate_constrainable_elements(
                        self._cib.xpath("//*[@id=$id]", id=res_id),
                        in_multiinstance_allowed,
                    ),
                    [
                        ReportItemFixture(
                            severity,
                            reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                            dict(
                                resource_id=res_id,
                                parent_type=(
                                    "bundle" if res_id == "B_R" else "clone"
                                ),
                                parent_id=res_id.split("_", 1)[0],
                            ),
                            force_code,
                            context=None,
                        )
                    ],
                )

    def test_error(self):
        self._test_report(
            False, reports.ReportItemSeverity.ERROR, reports.codes.FORCE
        )

    def test_warning(self):
        self._test_report(True, reports.ReportItemSeverity.WARNING, None)


class CreateConstraintWithSet(TestCase):
    def setUp(self):
        cib = str_to_etree(
            """
            <cib>
                <resources>
                    <primitive id="resource1" />
                    <primitive id="resource2" />
                    <primitive id="resource3" />
                    <primitive id="resource4" />
                    <primitive id="resource5" />
                    <primitive id="resource6" />
                </resources>
                <constraints />
            </cib>
            """
        )
        self.el_constraints = cib.find("constraints")
        self.id_provider = IdProvider(cib)

    @staticmethod
    def set(*args, **kwargs):
        return dict(ids=args, options=kwargs)

    def test_no_options(self):
        create_constraint_with_set(
            self.el_constraints,
            self.id_provider,
            Version(3, 7, 0),
            "rsc_constr",
            "constr",
            [
                self.set("resource1"),
                self.set("resource2", "resource3"),
                self.set("resource4", "resource5", "resource6"),
            ],
            {},
        )
        assert_xml_equal(
            """
            <constraints>
                <rsc_constr id="constr_set_r1r2r3">
                    <resource_set id="constr_set_r1r2r3_set">
                        <resource_ref id="resource1" />
                    </resource_set>
                    <resource_set id="constr_set_r1r2r3_set-1">
                        <resource_ref id="resource2" />
                        <resource_ref id="resource3" />
                    </resource_set>
                    <resource_set id="constr_set_r1r2r3_set-2">
                        <resource_ref id="resource4" />
                        <resource_ref id="resource5" />
                        <resource_ref id="resource6" />
                    </resource_set>
                </rsc_constr>
            </constraints>
            """,
            etree_to_str(self.el_constraints),
        )

    def test_options(self):
        create_constraint_with_set(
            self.el_constraints,
            self.id_provider,
            Version(3, 7, 0),
            "rsc_constr",
            "constr",
            [
                self.set("resource1", id="my-set"),
                self.set(
                    "resource2", "resource3", role=const.PCMK_ROLE_PROMOTED
                ),
            ],
            {"id": "id-test", "opt1": "val1", "opt2": "val2", "empty": ""},
        )
        assert_xml_equal(
            f"""
            <constraints>
                <rsc_constr id="id-test" opt1="val1" opt2="val2">
                    <resource_set id="my-set">
                        <resource_ref id="resource1" />
                    </resource_set>
                    <resource_set id="id-test_set"
                        role="{const.PCMK_ROLE_PROMOTED}"
                    >
                        <resource_ref id="resource2" />
                        <resource_ref id="resource3" />
                    </resource_set>
                </rsc_constr>
            </constraints>
            """,
            etree_to_str(self.el_constraints),
        )

    def test_legacy_role(self):
        create_constraint_with_set(
            self.el_constraints,
            self.id_provider,
            Version(3, 6, 0),
            "rsc_constr",
            "constr",
            [self.set("resource1", role=const.PCMK_ROLE_PROMOTED)],
            {},
        )
        assert_xml_equal(
            f"""
            <constraints>
                <rsc_constr id="constr_set_r1">
                    <resource_set id="constr_set_r1_set"
                        role="{const.PCMK_ROLE_PROMOTED_LEGACY}"
                    >
                        <resource_ref id="resource1" />
                    </resource_set>
                </rsc_constr>
            </constraints>
            """,
            etree_to_str(self.el_constraints),
        )



from unittest import TestCase, mock

from pcs.cli.alert.command import alert_config
from pcs.cli.common.errors import CmdLineInputError
from pcs.common.pacemaker.alert import (
    CibAlertDto,
    CibAlertListDto,
    CibAlertRecipientDto,
    CibAlertSelectAttributeDto,
    CibAlertSelectDto,
)
from pcs.common.pacemaker.nvset import CibNvpairDto, CibNvsetDto

from pcs_test.tools.misc import dict_to_modifiers

# The fixtures are almost the same as in tier1.test_alert and they should be
# unified and deduplicated eventually. Currently, it is not possible, because
# the tier0 test tests select, which is not supported by code tested in tier1
# test.
FIXTURE_ALERTS_DTO = CibAlertListDto(
    [
        CibAlertDto(
            id="alert1",
            path="/path/1",
            description=None,
            recipients=[],
            select=None,
            meta_attributes=[],
            instance_attributes=[],
        ),
        CibAlertDto(
            id="alert2",
            path="/path/2",
            description=None,
            recipients=[
                CibAlertRecipientDto(
                    id="alert2-recipient1",
                    value="test_value_1",
                    description=None,
                    meta_attributes=[],
                    instance_attributes=[],
                ),
                CibAlertRecipientDto(
                    id="alert2-recipient2",
                    value="test_value_2",
                    description="alert2 recipient2 description",
                    meta_attributes=[],
                    instance_attributes=[],
                ),
            ],
            select=None,
            meta_attributes=[],
            instance_attributes=[],
        ),
        CibAlertDto(
            id="alert-all",
            path="/path/all",
            description="alert all options",
            recipients=[
                CibAlertRecipientDto(
                    id="alert-all-recipient",
                    value="value-all",
                    description="all options recipient",
                    meta_attributes=[
                        CibNvsetDto(
                            id="aar1m",
                            options={},
                            rule=None,
                            nvpairs=[
                                CibNvpairDto(
                                    id="aar1m1", name="aar1m1n", value="aar1m1v"
                                ),
                                CibNvpairDto(
                                    id="aar1m2", name="aar1m2n", value="aar1m2v"
                                ),
                            ],
                        )
                    ],
                    instance_attributes=[
                        CibNvsetDto(
                            id="aar1i",
                            options={},
                            rule=None,
                            nvpairs=[
                                CibNvpairDto(
                                    id="aar1i1", name="aar1i1n", value="aar1i1v"
                                ),
                                CibNvpairDto(
                                    id="aar1i2", name="aar1i2n", value="aar1i2v"
                                ),
                            ],
                        )
                    ],
                )
            ],
            select=CibAlertSelectDto(
                nodes=True,
                fencing=False,
                resources=False,
                attributes=True,
                attributes_select=[
                    CibAlertSelectAttributeDto(id="attr1", name="standby"),
                    CibAlertSelectAttributeDto(id="attr2", name="shutdown"),
                ],
            ),
            meta_attributes=[
                CibNvsetDto(
                    id="aam1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="aam1", name="aam1n", value="aam1v"),
                        CibNvpairDto(id="aam2", name="aam2n", value="aam2v"),
                    ],
                )
            ],
            instance_attributes=[
                CibNvsetDto(
                    id="aai1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="aai1", name="aai1n", value="aai1v"),
                        CibNvpairDto(id="aai2", name="aai2n", value="aai2v"),
                    ],
                )
            ],
        ),
    ]
)

FIXTURE_ALERTS_TEXT = """\
Alert: alert1
  Path: /path/1
Alert: alert2
  Path: /path/2
  Recipients:
    Recipient: alert2-recipient1
      Value: test_value_1
    Recipient: alert2-recipient2
      Description: alert2 recipient2 description
      Value: test_value_2
Alert: alert-all
  Description: alert all options
  Path: /path/all
  Recipients:
    Recipient: alert-all-recipient
      Description: all options recipient
      Value: value-all
      Attributes: aar1i
        aar1i1n=aar1i1v
        aar1i2n=aar1i2v
      Meta Attributes: aar1m
        aar1m1n=aar1m1v
        aar1m2n=aar1m2v
  Receives:
    nodes
    attributes: 'shutdown', 'standby'
  Attributes: aai1
    aai1n=aai1v
    aai2n=aai2v
  Meta Attributes: aam1
    aam1n=aam1v
    aam2n=aam2v\
"""

FIXTURE_ALERTS_CMD = """\
pcs -- alert create path=/path/1 id=alert1;
pcs -- alert create path=/path/2 id=alert2;
pcs -- alert recipient add alert2 value=test_value_1 id=alert2-recipient1;
pcs -- alert recipient add alert2 value=test_value_2 id=alert2-recipient2 description='alert2 recipient2 description';
pcs -- alert create path=/path/all id=alert-all description='alert all options' options aai1n=aai1v aai2n=aai2v meta aam1n=aam1v aam2n=aam2v;
pcs -- alert recipient add alert-all value=value-all id=alert-all-recipient description='all options recipient' options aar1i1n=aar1i1v aar1i2n=aar1i2v meta aar1m1n=aar1m1v aar1m2n=aar1m2v\
"""


class AlertsConfigBaseMixin:
    def setUp(self):
        self.lib = mock.Mock(spec_set=["alert"])
        self.lib.alert = mock.Mock(spec_set=["get_config_dto"])
        self.lib_cmd = self.lib.alert.get_config_dto

    def _call_cmd(self, argv, modifiers=None):
        modifiers = modifiers or {}
        alert_config(self.lib, argv, dict_to_modifiers(modifiers))

    def test_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["arg"])
        self.assertIsNone(cm.exception.message)
        self.lib_cmd.assert_not_called()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.alert.command.print")
class AlertsConfigText(AlertsConfigBaseMixin, TestCase):
    def test_empty(self, mock_print):
        self.lib_cmd.return_value = CibAlertListDto([])
        self._call_cmd([])
        mock_print.assert_not_called()

    def test_complex(self, mock_print):
        self.lib_cmd.return_value = FIXTURE_ALERTS_DTO
        self._call_cmd([])
        mock_print.assert_called_once_with(FIXTURE_ALERTS_TEXT)


@mock.patch("pcs.cli.alert.command.print")
class AlertsConfigCmd(AlertsConfigBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None):
        modifiers = modifiers or {}
        modifiers["output-format"] = "cmd"
        return super()._call_cmd(argv, modifiers)

    def test_empty(self, mock_print):
        self.lib_cmd.return_value = CibAlertListDto([])
        self._call_cmd([])
        mock_print.assert_not_called()

    def test_complex(self, mock_print):
        self.lib_cmd.return_value = FIXTURE_ALERTS_DTO
        self._call_cmd([])
        mock_print.assert_called_once_with(FIXTURE_ALERTS_CMD)



from typing import Iterable

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.types import StringCollection
from pcs.lib.cib import element_description as cib_element_description
from pcs.lib.cib.remove_elements import (
    ElementsToRemove,
    ensure_resources_stopped,
    remove_specified_elements,
)
from pcs.lib.cib.resource.guest_node import is_guest_node
from pcs.lib.cib.resource.remote_node import (
    get_node_name_from_resource as get_node_name_from_remote_resource,
)
from pcs.lib.cib.resource.stonith import is_stonith
from pcs.lib.cib.tools import ElementNotFound, get_element_by_id, get_resources
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.sbd_stonith import ensure_some_stonith_remains


def remove_elements(
    env: LibraryEnvironment,
    ids: StringCollection,
    force_flags: reports.types.ForceFlags = (),
) -> None:
    """
    Remove elements with specified ids from CIB. This function is aware of
    relations and references between elements and will also remove all elements
    that are somehow referencing elements with specified ids.

    ids -- ids of configuration elements to remove
    force_flags -- list of flags codes
    """
    cib = env.get_cib()
    report_processor = env.report_processor

    elements_to_remove = ElementsToRemove(cib, ids)

    stonith_ids = []
    non_stonith_ids = []
    for res_el in elements_to_remove.resources_to_remove:
        if is_stonith(res_el):
            stonith_ids.append(str(res_el.attrib["id"]))
        else:
            non_stonith_ids.append(str(res_el.attrib["id"]))

    report_processor.report_list(
        _validate_elements_to_remove(elements_to_remove)
        + _ensure_not_guest_remote(elements_to_remove.resources_to_remove)
        + ensure_some_stonith_remains(
            env,
            get_resources(cib),
            stonith_resources_to_ignore=stonith_ids,
            sbd_being_disabled=False,
            force_flags=force_flags,
        )
    )

    # stonith resources do not need to be stopped, because they do not left
    # orphaned resources behind, so we check only the state of non-stonith
    # resources
    if non_stonith_ids:
        if env.is_cib_live:
            report_processor.report_list(
                ensure_resources_stopped(
                    env.get_cluster_state(), non_stonith_ids, force_flags
                )
            )
        else:
            report_processor.report(
                reports.ReportItem.warning(
                    reports.messages.StoppedResourcesBeforeDeleteCheckSkipped(
                        non_stonith_ids, reports.const.REASON_NOT_LIVE_CIB
                    )
                )
            )
    if report_processor.has_errors:
        raise LibraryError()

    report_processor.report_list(
        elements_to_remove.dependant_elements.to_reports()
    )
    report_processor.report_list(
        elements_to_remove.element_references.to_reports()
    )
    remove_specified_elements(cib, elements_to_remove)
    env.push_cib()


def element_description_set(
    env: LibraryEnvironment, element_id: str, description: str
) -> None:
    """
    Set, update, or delete the description of specified element.

    env -- LibraryEnvironment
    element_id -- id of the element
    description -- new description, if empty, delete existing description
    """
    try:
        element = get_element_by_id(env.get_cib(), element_id)
        env.report_processor.report_list(
            cib_element_description.validate_description_support(element)
        )
    except ElementNotFound:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.IdNotFound(element_id, [])
            )
        )
    if env.report_processor.has_errors:
        raise LibraryError()

    cib_element_description.set_description(element, description)
    env.push_cib()


def element_description_get(env: LibraryEnvironment, element_id: str) -> str:
    """
    Get description of specified element.

    env -- LibraryEnvironment
    element_id -- id of the element
    """
    try:
        element = get_element_by_id(env.get_cib(), element_id)
        env.report_processor.report_list(
            cib_element_description.validate_description_support(element)
        )
    except ElementNotFound:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.IdNotFound(element_id, [])
            )
        )
    if env.report_processor.has_errors:
        raise LibraryError()

    return cib_element_description.get_description(element)


def _validate_elements_to_remove(
    element_to_remove: ElementsToRemove,
) -> reports.ReportItemList:
    report_list = [
        reports.ReportItem.error(reports.messages.IdNotFound(missing_id, []))
        for missing_id in sorted(element_to_remove.missing_ids)
    ]
    unsupported_elements = element_to_remove.unsupported_elements
    report_list.extend(
        reports.ReportItem.error(
            reports.messages.IdBelongsToUnexpectedType(
                unsupported_id,
                list(unsupported_elements.supported_element_types),
                unsupported_elements.id_tag_map[unsupported_id],
            )
        )
        for unsupported_id in sorted(unsupported_elements.id_tag_map)
    )
    return report_list


def _ensure_not_guest_remote(
    resource_elements: Iterable[_Element],
) -> reports.ReportItemList:
    report_list = []
    for element in resource_elements:
        if is_guest_node(element):
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.UseCommandNodeRemoveGuest(
                        str(element.attrib["id"])
                    )
                )
            )
        if get_node_name_from_remote_resource(element) is not None:
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.UseCommandNodeRemoveRemote(
                        str(element.attrib["id"])
                    )
                )
            )
    return report_list
